"""
Tests for bugs fixed in v2.2.1:
1. DOT graph export - incorrect variable reference
2. Race condition in async lazy proxy
3. Silent cleanup failures now logged
"""
import asyncio
import logging
import pytest
import tempfile
import os
from unittest.mock import MagicMock, patch

from pico_ioc import init, component, configure, cleanup
from pico_ioc.aop import UnifiedComponentProxy
from pico_ioc.scope import ScopedCaches, ScopeManager, ComponentContainer
from pico_ioc.constants import PICO_META


# Define components at module level for proper scanning
@component()
class GraphServiceA:
    pass


@component()
class GraphServiceB:
    def __init__(self, a: GraphServiceA):
        self.a = a


@component()
class GraphDatabase:
    pass


@component()
class GraphCache:
    pass


@component()
class GraphRepository:
    def __init__(self, db: GraphDatabase, cache: GraphCache):
        self.db = db
        self.cache = cache


@component()
class GraphService:
    def __init__(self, repo: GraphRepository):
        self.repo = repo


class TestDotGraphExportFix:
    """Test that export_graph produces correct DOT syntax."""

    def test_dot_graph_edges_use_correct_variable(self):
        """Verify edges use {cid} not {child} - the v2.2.1 fix."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path)

            with open(path, 'r') as f:
                dot_output = f.read()

            # Should contain valid edge syntax
            assert "->" in dot_output
            # Should NOT contain literal {child} which was the bug
            assert "{child}" not in dot_output
            # Should reference actual node IDs
            assert "GraphServiceA" in dot_output
            assert "GraphServiceB" in dot_output

            os.unlink(path)
        finally:
            container.shutdown()

    def test_dot_graph_with_multiple_dependencies(self):
        """Test DOT export with complex dependency tree."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path)

            with open(path, 'r') as f:
                dot_output = f.read()

            # Count edges - should have multiple
            edge_count = dot_output.count("->")
            assert edge_count >= 3

            # Verify no malformed syntax
            lines = dot_output.split('\n')
            for line in lines:
                if '->' in line:
                    # Each edge line should not have unresolved variables
                    assert '{child}' not in line
                    assert '{cid}' not in line

            os.unlink(path)
        finally:
            container.shutdown()


class TestAsyncProxyRaceConditionFix:
    """Test the race condition fix in _async_init_if_needed."""

    @pytest.mark.asyncio
    async def test_concurrent_async_init_creates_single_instance(self):
        """Multiple concurrent awaits should create only one instance."""
        creation_count = 0

        def creator():
            nonlocal creation_count
            creation_count += 1
            return {"id": creation_count}

        container = MagicMock()
        container._run_configure_methods = MagicMock(return_value=None)

        proxy = UnifiedComponentProxy(
            container=container,
            target=None,
            object_creator=creator,
            component_key="test"
        )

        # Simulate concurrent access
        async def access_proxy():
            await proxy._async_init_if_needed()
            return proxy._get_real_object()

        # Run multiple concurrent accesses
        results = await asyncio.gather(*[access_proxy() for _ in range(10)])

        # All results should be the same object
        first = results[0]
        for r in results[1:]:
            assert r is first

        # Creator should only be called once
        assert creation_count == 1

    @pytest.mark.asyncio
    async def test_async_init_with_async_configure(self):
        """Test that async configure methods are properly awaited."""
        configure_called = False

        def creator():
            return MagicMock()

        async def async_configure_result():
            nonlocal configure_called
            await asyncio.sleep(0.01)
            configure_called = True

        container = MagicMock()
        # Return an awaitable from _run_configure_methods
        container._run_configure_methods = MagicMock(return_value=async_configure_result())

        proxy = UnifiedComponentProxy(
            container=container,
            target=None,
            object_creator=creator,
            component_key="test"
        )

        await proxy._async_init_if_needed()

        # Configure should have been awaited
        assert configure_called

    @pytest.mark.asyncio
    async def test_async_init_skips_if_already_initialized(self):
        """Fast path: skip initialization if target already exists."""
        creator = MagicMock(return_value={"test": True})

        container = MagicMock()

        proxy = UnifiedComponentProxy(
            container=container,
            target={"existing": True},  # Already has target
            object_creator=creator,
            component_key="test"
        )

        await proxy._async_init_if_needed()

        # Creator should NOT be called since target already exists
        creator.assert_not_called()


class TestCleanupLoggingFix:
    """Test that cleanup failures are now logged instead of silently ignored."""

    def test_cleanup_object_logs_exception(self, caplog):
        """Verify cleanup method exceptions are logged at WARNING level."""

        class FailingService:
            def close(self):
                raise RuntimeError("Connection refused")

        # Mark close as a cleanup method
        FailingService.close._pico_meta = {"cleanup": True}

        caches = ScopedCaches()
        service = FailingService()

        with caplog.at_level(logging.WARNING, logger="pico_ioc.scope"):
            caches._cleanup_object(service)

        # Should log the failure
        assert any("Cleanup method" in record.message for record in caplog.records)
        assert any("failed" in record.message.lower() for record in caplog.records)

    def test_cleanup_container_processes_all_items(self):
        """Multiple objects in container should all be processed."""
        caches = ScopedCaches()
        container = ComponentContainer()

        # Add multiple items
        container.put("s1", {"name": "service1"})
        container.put("s2", {"name": "service2"})

        cleanup_calls = []
        original_cleanup = caches._cleanup_object

        def tracking_cleanup(obj):
            cleanup_calls.append(obj)
            original_cleanup(obj)

        caches._cleanup_object = tracking_cleanup

        caches._cleanup_container(container)

        # Both objects should have been processed
        assert len(cleanup_calls) == 2

    def test_cleanup_scope_removes_entry_and_cleans(self):
        """cleanup_scope should remove bucket entry and run cleanup."""
        caches = ScopedCaches()
        scope_name = "request"
        scope_id = "req-123"

        # Manually add a scoped container
        container = ComponentContainer()
        container.put("key", {"data": "test"})
        caches._by_scope[scope_name] = {scope_id: container}

        # Cleanup should remove the entry
        caches.cleanup_scope(scope_name, scope_id)

        # Entry should be removed
        assert scope_id not in caches._by_scope.get(scope_name, {})


class TestScopedCachesEdgeCases:
    """Additional tests for ScopedCaches coverage."""

    def test_for_scope_singleton_returns_same_container(self):
        """Singleton scope always returns the same container."""
        caches = ScopedCaches()
        scopes = ScopeManager()

        c1 = caches.for_scope(scopes, "singleton")
        c2 = caches.for_scope(scopes, "singleton")

        assert c1 is c2

    def test_for_scope_prototype_returns_no_cache(self):
        """Prototype scope returns a no-op container."""
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = caches.for_scope(scopes, "prototype")

        # Should be the _NoCacheContainer
        container.put("key", "value")
        assert container.get("key") is None  # No caching
        assert container.items() == []

    def test_all_items_iterates_all_caches(self):
        """all_items should yield from singleton and all scoped caches."""
        caches = ScopedCaches()

        # Add to singleton
        caches._singleton.put("singleton_key", "singleton_val")

        # Add to scoped
        scoped_container = ComponentContainer()
        scoped_container.put("scoped_key", "scoped_val")
        caches._by_scope["request"] = {"req-1": scoped_container}

        items = list(caches.all_items())

        assert ("singleton_key", "singleton_val") in items
        assert ("scoped_key", "scoped_val") in items

    def test_shrink_does_nothing_for_singleton(self):
        """shrink() should be no-op for singleton scope."""
        caches = ScopedCaches()
        caches._singleton.put("key", "value")

        caches.shrink("singleton", 0)

        # Singleton should be unaffected
        assert caches._singleton.get("key") == "value"

    def test_shrink_does_nothing_for_prototype(self):
        """shrink() should be no-op for prototype scope."""
        caches = ScopedCaches()

        # Should not raise
        caches.shrink("prototype", 0)

    def test_shrink_removes_oldest_entries(self):
        """shrink() should remove entries beyond keep limit."""
        caches = ScopedCaches()

        # Add multiple scope entries
        for i in range(5):
            container = ComponentContainer()
            container.put(f"key_{i}", f"value_{i}")
            caches._by_scope.setdefault("request", {})[f"req-{i}"] = container

        assert len(caches._by_scope["request"]) == 5

        # Shrink to keep only 2
        caches.shrink("request", 2)

        assert len(caches._by_scope["request"]) == 2


class TestScopeManagerEdgeCases:
    """Additional tests for ScopeManager coverage."""

    def test_get_id_returns_none_for_singleton(self):
        """get_id for singleton always returns None."""
        sm = ScopeManager()
        assert sm.get_id("singleton") is None

    def test_get_id_returns_none_for_prototype(self):
        """get_id for prototype always returns None."""
        sm = ScopeManager()
        assert sm.get_id("prototype") is None

    def test_activate_returns_none_for_singleton(self):
        """activate() for singleton is no-op."""
        sm = ScopeManager()
        result = sm.activate("singleton", "any-id")
        assert result is None

    def test_deactivate_does_nothing_for_singleton(self):
        """deactivate() for singleton is no-op."""
        sm = ScopeManager()
        # Should not raise
        sm.deactivate("singleton", None)

    def test_names_excludes_reserved(self):
        """names() should not include singleton or prototype."""
        sm = ScopeManager()
        names = sm.names()

        assert "singleton" not in names
        assert "prototype" not in names
        assert "request" in names
        assert "session" in names

    def test_signature_returns_tuple_of_ids(self):
        """signature() returns tuple of scope IDs for given names."""
        sm = ScopeManager()

        # Activate some scopes
        sm.activate("request", "req-1")
        sm.activate("session", "sess-1")

        sig = sm.signature(("request", "session"))

        assert sig == ("req-1", "sess-1")

    def test_signature_all_returns_all_scope_ids(self):
        """signature_all() returns IDs for all registered scopes."""
        sm = ScopeManager()

        sm.activate("request", "req-1")

        sig = sm.signature_all()

        # Should have entries for all scopes (even if None)
        assert len(sig) == len(sm.names())



# tests/test_event_bus.py
import asyncio
import threading
import pytest
from typing import List
from pico_ioc import EventBus, ExecPolicy, ErrorPolicy, Event, subscribe, AutoSubscriberMixin

class MyEvent(Event):
    def __init__(self, value: int):
        self.value = value

def test_subscribe_and_publish_sync():
    bus = EventBus()
    seen: List[int] = []
    def handler(evt: MyEvent):
        seen.append(evt.value)
    bus.subscribe(MyEvent, handler)
    bus.publish_sync(MyEvent(42))
    assert seen == [42]

def test_priority_order_sync():
    bus = EventBus()
    seen: List[str] = []
    def h_low(evt: MyEvent):
        seen.append("low")
    def h_high(evt: MyEvent):
        seen.append("high")
    bus.subscribe(MyEvent, h_low, priority=0)
    bus.subscribe(MyEvent, h_high, priority=10)
    bus.publish_sync(MyEvent(1))
    assert seen == ["high", "low"]

@pytest.mark.asyncio
async def test_once_subscription_and_unsubscribe():
    bus = EventBus()
    seen: List[int] = []
    def handler(evt: MyEvent):
        seen.append(evt.value)
    bus.subscribe(MyEvent, handler, once=True)
    await bus.publish(MyEvent(1))
    await bus.publish(MyEvent(2))
    assert seen == [1]

@pytest.mark.asyncio
async def test_async_handler_inline_and_task_policy():
    bus = EventBus()
    seen: List[str] = []
    async def h_inline(evt: MyEvent):
        await asyncio.sleep(0)
        seen.append("inline")
    async def h_task(evt: MyEvent):
        await asyncio.sleep(0)
        seen.append("task")
    bus.subscribe(MyEvent, h_inline, priority=0, policy=ExecPolicy.INLINE)
    bus.subscribe(MyEvent, h_task, priority=0, policy=ExecPolicy.TASK)
    await bus.publish(MyEvent(0))
    assert set(seen) == {"inline", "task"}

@pytest.mark.asyncio
async def test_threadpool_policy():
    bus = EventBus()
    seen: List[str] = []
    def h_tp(evt: MyEvent):
        seen.append("tp")
    bus.subscribe(MyEvent, h_tp, policy=ExecPolicy.THREADPOOL)
    await bus.publish(MyEvent(0))
    assert seen == ["tp"]

@pytest.mark.asyncio
async def test_worker_queue_post_from_thread():
    bus = EventBus(max_queue_size=10)
    seen: List[int] = []
    def handler(evt: MyEvent):
        seen.append(evt.value)
    bus.subscribe(MyEvent, handler)
    await bus.start_worker()
    def producer():
        for i in range(3):
            bus.post(MyEvent(i))
    t = threading.Thread(target=producer)
    t.start()
    t.join()
    await asyncio.sleep(0.05)
    await bus.stop_worker()
    assert seen == [0, 1, 2]

@pytest.mark.asyncio
async def test_error_policy_raise():
    bus = EventBus(error_policy=ErrorPolicy.RAISE)
    def bad(evt: MyEvent):
        raise RuntimeError("boom")
    bus.subscribe(MyEvent, bad)
    with pytest.raises(Exception):
        await bus.publish(MyEvent(0))

def test_auto_subscriber_mixin():
    seen: List[int] = []
    class S(AutoSubscriberMixin):
        @subscribe(MyEvent, priority=5)
        def handle(self, evt: MyEvent):
            seen.append(evt.value)
    bus = EventBus()
    s = S()
    s._pico_autosubscribe(bus)
    bus.publish_sync(MyEvent(7))
    assert seen == [7]




import pytest
import os
import types
import asyncio
import pickle
import logging
from dataclasses import dataclass
from typing import List, Any, Callable, Optional
import pico_ioc.event_bus
from pico_ioc import (
    init, component, factory, provides, configured, configuration,
    cleanup, configure, health, intercepted_by,
    PicoContainer, MethodInterceptor, MethodCtx, ScopeError,
    InvalidBindingError, ComponentCreationError,
    ProviderNotFoundError, ConfigurationError, SerializationError,
    Event, subscribe, AutoSubscriberMixin, EventBus, EventBusClosedError
)

log_capture = []

class ListLogHandler(logging.Handler):
    def emit(self, record):
        log_capture.append(self.format(record))

test_logger = logging.getLogger("TestCoverageLogger")
test_logger.handlers.clear()
test_logger.addHandler(ListLogHandler())
test_logger.setLevel(logging.INFO)

@pytest.fixture(autouse=True)
def reset_logging_capture():
    log_capture.clear()

@component
class MissingDependency:
    pass

@component
class NeedsMissingComponent:
    def __init__(self, missing: MissingDependency):
        self.missing = missing

@component
class CircularA:
    def __init__(self, b: "CircularB"):
        self.b = b

@component
class CircularB:
    def __init__(self, a: CircularA):
        self.a = a

@component
class FailingComponent:
    def __init__(self):
        raise ValueError("Creation failure")

@configured(target="self", mapping="auto")
@dataclass
class RequiredConfig:
    REQUIRED_KEY: str

class Widget:
    pass

@factory
class WidgetFactory:
    @provides(Widget, primary=True)
    def build_widget(self) -> Widget:
        test_logger.info("Widget_Factory: Creating Widget")
        return Widget()

@component(lazy=True)
class ConfiguredComponent:
    def __init__(self):
        self.configured = False
        self.widget = None
        test_logger.info("ConfiguredComponent: __init__")
    @configure
    def setup(self, widget: Widget):
        self.widget = widget
        self.configured = True
        test_logger.info("ConfiguredComponent: @configure called")

@component
class AsyncResource:
    def __init__(self):
        self.closed = False
        test_logger.info("AsyncResource: Created")
    @cleanup
    async def async_close(self):
        await asyncio.sleep(0)
        self.closed = True
        test_logger.info("AsyncResource: @cleanup async called")

@component(conditional_require_env=("MY_TEST_VAR",))
class EnvConditionalComponent:
    pass

def check_predicate():
    return os.environ.get("PREDICATE_SWITCH") == "ON"

@component(conditional_predicate=check_predicate)
class PredicateConditionalComponent:
    pass

@component
class HealthyComponent:
    @health
    def check_db(self):
        test_logger.info("Health: check_db OK")
        return True
    @health
    def check_api(self):
        test_logger.info("Health: check_api FAILED")
        raise ValueError("API not available")

@component
class AuditInterceptor(MethodInterceptor):
    def invoke(self, ctx: MethodCtx, call_next: Callable[[MethodCtx], Any]) -> Any:
        test_logger.info(f"AUDIT_ASYNC - Entering: {ctx.name}")
        res = call_next(ctx)
        test_logger.info(f"AUDIT_ASYNC - Exiting: {ctx.name}")
        return res

@component
class AsyncAuditedService:
    @intercepted_by(AuditInterceptor)
    async def do_async_work(self, val: int) -> int:
        test_logger.info("AsyncAuditedService: working...")
        await asyncio.sleep(0)
        return val * 2

@component(lazy=True)
class MyLazyComponentForPickle:
    def __init__(self):
        self.value = 42

class MyTestEvent(Event):
    def __init__(self, msg: str):
        self.msg = msg

@component(lazy=True)
class MyTestSubscriber(AutoSubscriberMixin):
    def __init__(self):
        self.received: List[str] = []

    @subscribe(MyTestEvent)
    def handle_event(self, evt: MyTestEvent):
        test_logger.info(f"Event received: {evt.msg}")
        self.received.append(evt.msg)

test_module = types.ModuleType("test_coverage_module")
all_definitions = [
    NeedsMissingComponent, MissingDependency, CircularA, CircularB, FailingComponent,
    RequiredConfig, Widget, WidgetFactory, ConfiguredComponent, AsyncResource,
    EnvConditionalComponent, PredicateConditionalComponent, HealthyComponent,
    AuditInterceptor, AsyncAuditedService, MyLazyComponentForPickle,
    MyTestEvent, MyTestSubscriber
]
for item in all_definitions:
    if hasattr(item, "__name__"):
        setattr(test_module, item.__name__, item)
        if item is RequiredConfig:
            setattr(test_module, item.__name__, configured(target=item, mapping="auto")(item))


def test_invalid_binding_error_on_init():
    mod = types.ModuleType("fail_mod")
    setattr(mod, NeedsMissingComponent.__name__, NeedsMissingComponent)
    with pytest.raises(InvalidBindingError) as e:
        init(mod, validate_only=True)
    assert "depends on MissingDependency which is not bound" in str(e.value)

def test_circular_dependency_error():
    with pytest.raises(InvalidBindingError) as e:
        init(test_module)
    assert "CircularA -> CircularB -> CircularA" in str(e.value)

def test_component_creation_error():
    mod = types.ModuleType("fail_mod2")
    for cls in [FailingComponent, Widget, WidgetFactory, ConfiguredComponent]:
        setattr(mod, cls.__name__, cls)

    with pytest.raises(ComponentCreationError) as err:
        init(mod) 

    assert "FailingComponent" in str(err.value)
    assert "Creation failure" in str(err.value)

def test_provider_not_found_error():
    container = init(types.ModuleType("empty_mod"))
    with pytest.raises(ProviderNotFoundError) as e_str:
        container.get("non_existent_key")
    assert "Provider for key 'non_existent_key' not found" in str(e_str.value)
    class NonExistentClass:
        pass
    with pytest.raises(ProviderNotFoundError) as e_type:
        container.get(NonExistentClass)
    assert "Provider for key 'NonExistentClass' not found" in str(e_type.value)

def test_configuration_error_missing_value():
    config_module = types.ModuleType("config_test_mod")
    
    @configured(target="self", mapping="auto")
    @dataclass
    class LocalRequiredConfig:
        REQUIRED_KEY: str
    
    setattr(config_module, LocalRequiredConfig.__name__, LocalRequiredConfig)
    
    with pytest.raises(ComponentCreationError) as e:
        container = init(config_module, config=configuration())
        container.get(LocalRequiredConfig)
    assert isinstance(e.value.cause, ConfigurationError)
    assert "Missing configuration key: REQUIRED_KEY" in str(e.value.cause)

def test_configuration_error_disallowed_profile():
    with pytest.raises(ConfigurationError) as e:
        init(
            types.ModuleType("empty_mod"),
            profiles=("dev",),
            allowed_profiles=("prod", "test")
        )
    assert "Unknown profiles: ['dev']; allowed: ['prod', 'test']" in str(e.value)

def test_scope_error_unknown_scope():
    container = init(types.ModuleType("empty_mod"))
    with pytest.raises(ScopeError) as e:
        container.activate_scope("unreal_scope", "id-123")
    assert "Unknown scope: unreal_scope" in str(e.value)

def test_factory_and_provides_pattern():
    mod = types.ModuleType("factory_mod")
    for cls in [Widget, WidgetFactory]:
        setattr(mod, cls.__name__, cls)
    container = init(mod)
    widget_instance = container.get(Widget)
    assert isinstance(widget_instance, Widget)
    assert "Widget_Factory: Creating Widget" in log_capture

@pytest.mark.asyncio
async def test_configure_lifecycle_method():
    mod = types.ModuleType("configure_mod")
    for cls in [Widget, WidgetFactory, ConfiguredComponent]:
        setattr(mod, cls.__name__, cls)
    container = init(mod)

    await container.aget(ConfiguredComponent)

    assert "ConfiguredComponent: __init__" in log_capture
    assert "Widget_Factory: Creating Widget" in log_capture
    assert "ConfiguredComponent: @configure called" in log_capture

@pytest.mark.asyncio
async def test_async_cleanup_method():
    mod = types.ModuleType("cleanup_mod")
    for cls in [AsyncResource]:
        setattr(mod, cls.__name__, cls)
    container = init(mod)
    resource = container.get(AsyncResource)
    assert resource.closed is False
    assert "AsyncResource: Created" in log_capture
    await container.cleanup_all_async()
    assert resource.closed is True
    assert "AsyncResource: @cleanup async called" in log_capture

def test_conditional_on_env(monkeypatch):
    mod = types.ModuleType("env_mod")
    setattr(mod, EnvConditionalComponent.__name__, EnvConditionalComponent)
    monkeypatch.delenv("MY_TEST_VAR", raising=False)
    container_no_env = init(mod)
    with pytest.raises(ProviderNotFoundError):
        container_no_env.get(EnvConditionalComponent)
    monkeypatch.setenv("MY_TEST_VAR", "any_value")
    container_with_env = init(mod)
    instance = container_with_env.get(EnvConditionalComponent)
    assert isinstance(instance, EnvConditionalComponent)

def test_conditional_on_predicate(monkeypatch):
    mod = types.ModuleType("pred_mod")
    setattr(mod, PredicateConditionalComponent.__name__, PredicateConditionalComponent)
    monkeypatch.setenv("PREDICATE_SWITCH", "OFF")
    container_false = init(mod)
    with pytest.raises(ProviderNotFoundError):
        container_false.get(PredicateConditionalComponent)
    monkeypatch.setenv("PREDICATE_SWITCH", "ON")
    container_true = init(mod)
    instance = container_true.get(PredicateConditionalComponent)
    assert isinstance(instance, PredicateConditionalComponent)

def test_health_check_decorator():
    mod = types.ModuleType("health_mod")
    setattr(mod, HealthyComponent.__name__, HealthyComponent)
    container = init(mod)
    container.get(HealthyComponent)
    health_status = container.health_check()
    assert "HealthyComponent.check_db" in health_status
    assert health_status["HealthyComponent.check_db"] is True
    assert "HealthyComponent.check_api" in health_status
    assert health_status["HealthyComponent.check_api"] is False
    assert "Health: check_db OK" in log_capture
    assert "Health: check_api FAILED" in log_capture

@pytest.mark.asyncio
async def test_aop_on_async_method():
    mod = types.ModuleType("aop_mod")
    for cls in [AuditInterceptor, AsyncAuditedService]:
        setattr(mod, cls.__name__, cls)
    container = init(mod)
    service = container.get(AsyncAuditedService)
    result = await service.do_async_work(10)
    assert result == 20
    assert "AUDIT_ASYNC - Entering: do_async_work" in log_capture
    assert "AsyncAuditedService: working..." in log_capture
    assert "AUDIT_ASYNC - Exiting: do_async_work" in log_capture

def test_proxy_serialization_pickle():
    mod = types.ModuleType("pickle_mod")
    for cls in [MyLazyComponentForPickle, AuditInterceptor, AsyncAuditedService]:
        setattr(mod, cls.__name__, cls)
    container = init(mod)
    lazy_proxy = container.get(MyLazyComponentForPickle)
    intercepted_proxy = container.get(AsyncAuditedService)
    try:
        pickled_lazy = pickle.dumps(lazy_proxy)
        pickled_intercepted = pickle.dumps(intercepted_proxy)
        unpickled_lazy = pickle.loads(pickled_lazy)
        unpickled_intercepted = pickle.loads(pickled_intercepted)
        assert unpickled_lazy.value == 42
        assert isinstance(unpickled_intercepted, AsyncAuditedService)
    except (pickle.PicklingError, SerializationError) as e:
        pytest.fail(f"Proxy serialization failed: {e}")

@pytest.mark.asyncio
async def test_event_bus_integration_and_shutdown():
    mod = types.ModuleType("event_mod")
    for cls in [MyTestEvent, MyTestSubscriber]:
        setattr(mod, cls.__name__, cls)
    container = init([mod, pico_ioc.event_bus])
    bus = container.get(EventBus)
    subscriber = await container.aget(MyTestSubscriber)
    assert isinstance(bus, EventBus)
    assert len(subscriber.received) == 0
    assert "Event received: Hello" not in log_capture
    await bus.publish(MyTestEvent("Hello"))
    assert subscriber.received == ["Hello"]
    assert "Event received: Hello" in log_capture
    await container.cleanup_all_async()
    with pytest.raises(EventBusClosedError):
        await bus.publish(MyTestEvent("Goodbye"))

def test_dependency_graph_dot_export(tmp_path):
    import types
    from pico_ioc.api import component, provides, init, _format_key

    class Repo:
        pass

    class Service:
        pass

    class ServiceImpl(Service):
        def __init__(self, repo: Repo) -> None:
            self.repo = repo

    class Tool:
        pass

    def make_module():
        m = types.ModuleType("graph_mod_dot")

        @component
        class MyRepo(Repo):
            pass

        @provides(Service)
        def build_service(repo: Repo) -> Service:
            return ServiceImpl(repo)

        @provides(Tool)
        def build_tool(svc: Service) -> Tool:
            return Tool()

        setattr(m, "MyRepo", MyRepo)
        setattr(m, "build_service", build_service)
        setattr(m, "build_tool", build_tool)
        return m

    def graph_to_dot(graph):
        nodes = list(graph.keys())
        ids = {k: f"n{i}" for i, k in enumerate(nodes)}
        lines = []
        lines.append("digraph PicoIoC {")
        lines.append("rankdir=LR;")
        lines.append("node [shape=box];")
        for k in nodes:
            nid = ids[k]
            label = _format_key(k)
            lines.append(f'{nid} [label="{label}"];')
        for k, deps in graph.items():
            src = ids[k]
            for d in deps:
                if d in ids:
                    dst = ids[d]
                    lines.append(f"{src} -> {dst};")
        lines.append("}")
        return "\n".join(lines)

    mod = make_module()
    pico = init(mod, validate_only=True)
    graph = pico.build_resolution_graph()
    dot = graph_to_dot(graph)
    out = tmp_path / "dependencies.dot"
    out.write_text(dot, encoding="utf-8")
    assert "Service" in dot
    assert "Repo" in dot
    assert "Tool" in dot
    assert "->" in dot

def test_dependency_graph_includes_provides_functions():
    import types
    from pico_ioc.api import component, provides, init

    class Repo:
        pass

    class Service:
        pass

    class ServiceImpl(Service):
        def __init__(self, repo: Repo) -> None:
            self.repo = repo

    class Tool:
        pass

    def make_module():
        m = types.ModuleType("graph_mod")

        @component
        class MyRepo(Repo):
            pass

        @provides(Service)
        def build_service(repo: Repo) -> Service:
            return ServiceImpl(repo)

        @provides(Tool)
        def build_tool(svc: Service) -> Tool:
            return Tool()

        setattr(m, "MyRepo", MyRepo)
        setattr(m, "build_service", build_service)
        setattr(m, "build_tool", build_tool)
        return m

    mod = make_module()
    pico = init(mod, validate_only=True)
    graph = pico.build_resolution_graph()

    assert Service in graph
    assert Tool in graph
    assert any(getattr(k, "__name__", "") == "MyRepo" for k in graph)

    myrepo_key = next(k for k in graph if getattr(k, "__name__", "") == "MyRepo")
    assert myrepo_key in graph[Service]
    assert Service in graph[Tool]



from typing import Dict, List, Tuple, Union
from .factory import Provider, ProviderMetadata
from .config_registrar import ConfigurationManager

KeyT = Union[str, type]

class ProviderSelector:
    def __init__(self, config_manager: ConfigurationManager):
        self._config_manager = config_manager

    def _rank_provider(self, item: Tuple[bool, Provider, ProviderMetadata]) -> Tuple[int, int, int]:
        provider, md = item[1], item[2]
        
        is_present = 1 if self._config_manager.prefix_exists(md) else 0
        
        pref = str(md.pico_name or "")
        pref_len = len(pref)
        
        is_primary = 1 if item[0] else 0
        
        return (is_present, pref_len, is_primary)

    def select_providers(
        self,
        candidates: Dict[KeyT, List[Tuple[bool, Provider, ProviderMetadata]]],
    ) -> Dict[KeyT, Tuple[Provider, ProviderMetadata]]:
        
        winners: Dict[KeyT, Tuple[Provider, ProviderMetadata]] = {}
        
        for key, lst in candidates.items():
            lst_sorted = sorted(lst, key=self._rank_provider, reverse=True)
            chosen = lst_sorted[0]
            winners[key] = (chosen[1], chosen[2])
            
        return winners



import pytest
import pickle
import inspect
import threading
from pico_ioc.aop import (
    UnifiedComponentProxy, 
    intercepted_by, 
    MethodInterceptor, 
    MethodCtx,
    _gather_interceptors_for_method
)
from pico_ioc.exceptions import SerializationError, AsyncResolutionError

def test_intercepted_by_validation():
    with pytest.raises(TypeError, match="requires at least one"):
        intercepted_by()
        
    with pytest.raises(TypeError, match="expects interceptor classes"):
        intercepted_by("not_a_class")
        
    decorator = intercepted_by(MethodInterceptor)
    with pytest.raises(TypeError, match="only decorate callables"):
        decorator("not_callable")

def test_proxy_init_validation():
    with pytest.raises(ValueError, match="non-null container"):
        UnifiedComponentProxy(container=None)
        
    with pytest.raises(ValueError, match="requires either a target"):
        UnifiedComponentProxy(container="stub")

def test_proxy_creator_errors():
    proxy = UnifiedComponentProxy(container="stub", object_creator="not_callable")
    with pytest.raises(TypeError, match="must be callable"):
        getattr(proxy, "any_attr")

    proxy = UnifiedComponentProxy(container="stub", object_creator=lambda: None)
    with pytest.raises(RuntimeError, match="returned None"):
        getattr(proxy, "any_attr")

def test_proxy_async_error_on_sync_access():
    async def async_configure(obj):
        pass
        
    class ContainerMock:
        def _run_configure_methods(self, obj):
            return async_configure(obj)
            
    proxy = UnifiedComponentProxy(
        container=ContainerMock(), 
        object_creator=lambda: "target"
    )
    
    with pytest.raises(AsyncResolutionError):
        str(proxy)

def test_proxy_serialization_error():
    class Unpicklable:
        def __getstate__(self):
            raise ValueError("No pickle")
            
    proxy = UnifiedComponentProxy(container="stub", target=Unpicklable())
    
    with pytest.raises(SerializationError):
        pickle.dumps(proxy)
        
    with pytest.raises(SerializationError):
        proxy.__setstate__({"data": b"bad_data"})

def test_gather_interceptors_edge_cases():
    assert _gather_interceptors_for_method(str, "non_existent") == ()
    assert _gather_interceptors_for_method(list, "append") == ()



import pytest
from typing import (
    List, Set, Iterable, Deque, Dict, Mapping, Protocol, Any,
    Annotated, Type, cast, Sequence, Collection
)
from pico_ioc import (
    init, component, factory, provides, Qualifier,
    DependencyRequest, analyze_callable_dependencies
)

class IService(Protocol):
    def serve(self) -> str: ...

@component(name="serviceA", qualifiers=["q1"])
class ServiceA(IService):
    def serve(self) -> str:
        return "A"

@component(name="serviceB", qualifiers=["q2"])
class ServiceB(IService):
    def serve(self) -> str:
        return "B"

@component
class OtherComponent:
    pass

@component
class Consumer:
    def __init__(
        self,
        all_services_list: List[IService],
        all_services_set: Set[IService],
        all_services_iterable: Iterable[IService],
        str_map: Dict[str, IService],
        type_map: Dict[Type, IService],
        q1_list: Annotated[List[IService], Qualifier("q1")],
        others: List[OtherComponent]
    ):
        self.all_services_list = all_services_list
        self.all_services_set = all_services_set
        self.all_services_iterable = all_services_iterable
        self.str_map = str_map
        self.type_map = type_map
        self.q1_list = q1_list
        self.others = others

def test_analyze_collections_and_dicts():
    class Sample:
        def __init__(
            self,
            a: List[IService],
            b: Set[IService],
            c: Iterable[IService],
            d: Deque[IService],
            e: Dict[str, IService],
            f: Mapping[Type, IService],
            g: Annotated[List[IService], Qualifier("q")],
            h: Sequence[IService],
            i: Collection[IService]
        ):
            pass

    plan = analyze_callable_dependencies(Sample.__init__)
    
    assert len(plan) == 9
    
    assert plan[0].parameter_name == "a"
    assert plan[0].key == IService
    assert plan[0].is_list is True
    assert plan[0].is_dict is False
    
    assert plan[1].parameter_name == "b"
    assert plan[1].key == IService
    assert plan[1].is_list is True
    assert plan[1].is_dict is False

    assert plan[2].parameter_name == "c"
    assert plan[2].key == IService
    assert plan[2].is_list is True
    
    assert plan[3].parameter_name == "d"
    assert plan[3].key == IService
    assert plan[3].is_list is True

    assert plan[4].parameter_name == "e"
    assert plan[4].key == IService
    assert plan[4].is_list is False
    assert plan[4].is_dict is True
    assert plan[4].dict_key_type == str

    assert plan[5].parameter_name == "f"
    assert plan[5].key == IService
    assert plan[5].is_list is False
    assert plan[5].is_dict is True
    assert plan[5].dict_key_type == Type

    assert plan[6].parameter_name == "g"
    assert plan[6].key == IService
    assert plan[6].is_list is True
    assert plan[6].qualifier == "q"
    
    assert plan[7].parameter_name == "h"
    assert plan[7].key == IService
    assert plan[7].is_list is True

    assert plan[8].parameter_name == "i"
    assert plan[8].key == IService
    assert plan[8].is_list is True


def test_container_resolves_collections_and_dicts():
    
    container = init(modules=[__name__])
    consumer = container.get(Consumer)
    
    instance_a = container.get("serviceA")
    instance_b = container.get("serviceB")
    
    instance_other = container.get(OtherComponent)

    assert isinstance(consumer.all_services_list, list)
    assert len(consumer.all_services_list) == 2
    assert instance_a in consumer.all_services_list
    assert instance_b in consumer.all_services_list
    
    assert isinstance(consumer.all_services_set, list)
    assert len(consumer.all_services_set) == 2
    assert instance_a in consumer.all_services_set
    assert instance_b in consumer.all_services_set
    
    assert isinstance(consumer.all_services_iterable, list)
    assert len(consumer.all_services_iterable) == 2
    
    assert isinstance(consumer.str_map, dict)
    assert len(consumer.str_map) == 2
    assert consumer.str_map["serviceA"] is instance_a
    assert consumer.str_map["serviceB"] is instance_b
    
    assert isinstance(consumer.type_map, dict)
    assert len(consumer.type_map) == 2
    assert consumer.type_map[ServiceA] is instance_a
    assert consumer.type_map[ServiceB] is instance_b
    
    assert isinstance(consumer.q1_list, list)
    assert len(consumer.q1_list) == 1
    assert consumer.q1_list[0] is instance_a
    
    assert isinstance(consumer.others, list)
    assert len(consumer.others) == 1
    assert consumer.others[0] is instance_other



import os
import json
from dataclasses import dataclass
from typing import Any, Optional, Protocol, Mapping, List, Tuple, Dict, Union

from .config_runtime import TreeSource, DictSource, JsonTreeSource, YamlTreeSource, Value
from .exceptions import ConfigurationError

class ConfigSource(Protocol):
    pass

class EnvSource(ConfigSource):
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix
    def get(self, key: str) -> Optional[str]:
        return os.environ.get(self.prefix + key)

class FileSource(ConfigSource):
    def __init__(self, path: str, prefix: str = "") -> None:
        self.prefix = prefix
        try:
            with open(path, "r", encoding="utf-8") as f:
                self._data = json.load(f)
        except Exception:
            self._data = {}
    def get(self, key: str) -> Optional[str]:
        k = self.prefix + key
        v = self._data
        for part in k.split("__"):
            if isinstance(v, dict) and part in v:
                v = v[part]
            else:
                return None
        if isinstance(v, (str, int, float, bool)):
            return str(v)
        return None

class FlatDictSource(ConfigSource):
    def __init__(self, data: Mapping[str, Any], prefix: str = "", case_sensitive: bool = True):
        base = dict(data)
        if case_sensitive:
            self._data = {str(k): v for k, v in base.items()}
            self._prefix = prefix
        else:
            self._data = {str(k).upper(): v for k, v in base.items()}
            self._prefix = prefix.upper()
        self._case_sensitive = case_sensitive
    def get(self, key: str) -> Optional[str]:
        if not key:
            return None
        k = f"{self._prefix}{key}" if self._prefix else key
        if not self._case_sensitive:
            k = k.upper()
        v = self._data.get(k)
        if v is None:
            return None
        if isinstance(v, (str, int, float, bool)):
            return str(v)
        return None

@dataclass(frozen=True)
class ContextConfig:
    flat_sources: Tuple[Union[EnvSource, FileSource, FlatDictSource], ...]
    tree_sources: Tuple[TreeSource, ...]
    overrides: Dict[str, Any]

def configuration(
    *sources: Any,
    overrides: Optional[Dict[str, Any]] = None
) -> ContextConfig:
    
    flat: List[Union[EnvSource, FileSource, FlatDictSource]] = []
    tree: List[TreeSource] = []

    for src in sources:
        if isinstance(src, (EnvSource, FileSource, FlatDictSource)):
            flat.append(src)
        elif isinstance(src, TreeSource):
            tree.append(src)
        else:
            raise ConfigurationError(f"Unknown configuration source type: {type(src)}")
            
    return ContextConfig(
        flat_sources=tuple(flat),
        tree_sources=tuple(tree),
        overrides=dict(overrides or {})
    )



import logging

LOGGER_NAME = "pico_ioc"
LOGGER = logging.getLogger(LOGGER_NAME)

PICO_INFRA = "_pico_infra"
PICO_NAME = "_pico_name"
PICO_KEY = "_pico_key"
PICO_META = "_pico_meta"

SCOPE_SINGLETON = "singleton"
SCOPE_PROTOTYPE = "prototype"




import pytest
import json
import os
from dataclasses import dataclass
from typing import Mapping, Any

from pico_ioc.config_builder import (
    FlatDictSource,
    FileSource,
    configuration,
    EnvSource,
    ContextConfig,
    TreeSource
)
from pico_ioc.exceptions import ConfigurationError

class MockTreeSource(TreeSource):
    def get_tree(self) -> Mapping[str, Any]:
        return {}

def test_flat_dict_source_basic():
    data = {
        "APP_HOST": "localhost",
        "APP_PORT": 8080,
        "APP_DEBUG": True,
        "APP_SCORE": 99.9,
        "IGNORED_LIST": [1, 2]
    }
    source = FlatDictSource(data, prefix="APP_", case_sensitive=True)

    assert source.get("HOST") == "localhost"
    assert source.get("PORT") == "8080"
    assert source.get("DEBUG") == "True"
    assert source.get("SCORE") == "99.9"
    
    assert source.get("MISSING") is None
    
    source_no_prefix = FlatDictSource(data, prefix="")
    assert source_no_prefix.get("IGNORED_LIST") is None

def test_flat_dict_source_case_insensitive():
    data = {
        "app_host": "localhost",
        "APP_PORT": 8080
    }
    source = FlatDictSource(data, prefix="APP_", case_sensitive=False)

    assert source.get("host") == "localhost" 
    assert source.get("HOST") == "localhost"
    
    assert source.get("port") == "8080"

def test_flat_dict_source_edge_cases():
    source = FlatDictSource({"KEY": "val"}, prefix="")
    
    assert source.get("") is None
    assert source.get(None) is None

    assert source.get("KEY") == "val"

def test_file_source_exception_handling():
    source = FileSource("ruta/que/no/existe/config.json")
    
    assert source.get("cualquier_cosa") is None
    assert source._data == {}

def test_file_source_invalid_json(tmp_path):
    f = tmp_path / "bad.json"
    f.write_text("{ esto no es json }", encoding="utf-8")
    
    source = FileSource(str(f))
    assert source._data == {}

def test_file_source_traversal_and_types(tmp_path):
    data = {
        "section": {
            "key": 123,
            "deep": {
                "val": True
            }
        },
        "list_val": [1, 2, 3]
    }
    f = tmp_path / "config.json"
    f.write_text(json.dumps(data), encoding="utf-8")
    
    source = FileSource(str(f))

    assert source.get("section__key") == "123"
    assert source.get("section__deep__val") == "True"

    assert source.get("section") is None

    assert source.get("list_val") is None

    assert source.get("section__missing") is None

def test_configuration_builder_categorization():
    env_src = EnvSource()
    flat_src = FlatDictSource({})
    tree_src = MockTreeSource()

    config = configuration(env_src, flat_src, tree_src)

    assert isinstance(config, ContextConfig)
    assert env_src in config.flat_sources
    assert flat_src in config.flat_sources
    assert tree_src in config.tree_sources
    assert len(config.flat_sources) == 2
    assert len(config.tree_sources) == 1

def test_configuration_builder_unknown_type():
    class UnknownSource:
        pass

    with pytest.raises(ConfigurationError) as exc:
        configuration(UnknownSource())
    
    assert "Unknown configuration source type" in str(exc.value)



import os
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from .constants import LOGGER, PICO_INFRA, PICO_NAME, PICO_META, SCOPE_SINGLETON
from .factory import ComponentFactory, ProviderMetadata, DeferredProvider
from .locator import ComponentLocator
from .aop import UnifiedComponentProxy
from .config_builder import ContextConfig
from .config_runtime import TreeSource
from .config_registrar import ConfigurationManager
from .provider_selector import ProviderSelector
from .dependency_validator import DependencyValidator
from .component_scanner import ComponentScanner, CustomScanner
from .analysis import analyze_callable_dependencies, DependencyRequest
from .container import PicoContainer

KeyT = Union[str, type]
Provider = Callable[[], Any]

def _can_be_selected_for(reg_md: Dict[KeyT, ProviderMetadata], selector: Any) -> bool:
    if not isinstance(selector, type):
        return False
    for md in reg_md.values():
        typ = md.provided_type or md.concrete_class
        if isinstance(typ, type):
            try:
                if issubclass(typ, selector):
                    return True
            except Exception:
                continue
    return False

class Registrar:
    def __init__(self, factory: ComponentFactory, *, profiles: Tuple[str, ...] = (), environ: Optional[Dict[str, str]] = None, logger: Optional[logging.Logger] = None, config: Optional[ContextConfig] = None) -> None:
        self._factory = factory
        self._profiles = set(p.strip() for p in profiles if p)
        self._environ = environ if environ is not None else os.environ
        self._metadata: Dict[KeyT, ProviderMetadata] = {}
        self._indexes: Dict[str, Dict[Any, List[KeyT]]] = {}
        self._log = logger or LOGGER
        self._config_manager = ConfigurationManager(config)
        self._provider_selector = ProviderSelector(self._config_manager)
        self._scanner = ComponentScanner(self._profiles, self._environ, self._config_manager)
        self._deferred: List[DeferredProvider] = []
        self._provides_functions: Dict[KeyT, Callable[..., Any]] = {}

    def register_custom_scanner(self, scanner: CustomScanner) -> None:
        self._scanner.register_custom_scanner(scanner)

    def locator(self) -> ComponentLocator:
        loc = ComponentLocator(dict(self._metadata), dict(self._indexes))
        setattr(loc, "_provides_functions", dict(self._provides_functions))
        return loc

    def attach_runtime(self, pico, locator: ComponentLocator) -> None:
        for deferred in self._deferred:
            deferred.attach(pico, locator)
        for key, md in list(self._metadata.items()):
            if md.lazy:
                original = self._factory.get(key, origin='lazy')
                def lazy_proxy_provider(_orig=original, _p=pico, _k=key):
                    return UnifiedComponentProxy(container=_p, object_creator=_orig, component_key=_k)
                self._factory.bind(key, lazy_proxy_provider)

    def _bind_if_absent(self, key: KeyT, provider: Provider) -> None:
        if not self._factory.has(key):
            self._factory.bind(key, provider)

    def register_module(self, module: Any) -> None:
        self._scanner.scan_module(module)

    def _find_md_for_type(self, t: type) -> Optional[ProviderMetadata]:
        cands: List[ProviderMetadata] = []
        for md in self._metadata.values():
            typ = md.provided_type or md.concrete_class
            if not isinstance(typ, type):
                continue
            try:
                if issubclass(typ, t):
                    cands.append(md)
            except Exception:
                continue
        if not cands:
            return None
        prim = [m for m in cands if m.primary]
        return prim[0] if prim else cands[0]

    def _find_narrower_scope_from_deps(self, deps: Tuple[DependencyRequest, ...]) -> Optional[str]:
        if not deps:
            return None

        for dep_req in deps:
            if dep_req.is_list:
                continue

            dep_md = self._metadata.get(dep_req.key)
            if dep_md is None:
                if isinstance(dep_req.key, type):
                    dep_md = self._find_md_for_type(dep_req.key)

            if dep_md and dep_md.scope != SCOPE_SINGLETON:
                return dep_md.scope
        return None

    def _promote_scopes(self) -> None:
        for k, md in list(self._metadata.items()):
            if md.scope == SCOPE_SINGLETON:
                ns = self._find_narrower_scope_from_deps(md.dependencies)
                if ns and ns != SCOPE_SINGLETON:
                    self._metadata[k] = ProviderMetadata(key=md.key, provided_type=md.provided_type, concrete_class=md.concrete_class, factory_class=md.factory_class, factory_method=md.factory_method, qualifiers=md.qualifiers, primary=md.primary, lazy=md.lazy, infra=md.infra, pico_name=md.pico_name, override=md.override, scope=ns, dependencies=md.dependencies)

    def _rebuild_indexes(self) -> None:
        self._indexes.clear()
        def add(idx: str, val: Any, key: KeyT):
            b = self._indexes.setdefault(idx, {}).setdefault(val, [])
            if key not in b:
                b.append(key)

        for k, md in self._metadata.items():
            for q in md.qualifiers:
                add("qualifier", q, k)
            if md.primary:
                add("primary", True, k)
            add("lazy", bool(md.lazy), k)
            if md.infra is not None:
                add("infra", md.infra, k)
            if md.pico_name is not None:
                add("pico_name", md.pico_name, k)

    def finalize(self, overrides: Optional[Dict[KeyT, Any]], *, pico_instance: PicoContainer) -> None:
        candidates, on_missing, deferred_providers, provides_functions = self._scanner.get_scan_results()
        self._deferred = deferred_providers
        self._provides_functions = provides_functions

        winners = self._provider_selector.select_providers(candidates)
        for key, (provider, md) in winners.items():
            self._bind_if_absent(key, provider)
            self._metadata[key] = md
            
        if PicoContainer not in self._metadata:
            self._factory.bind(PicoContainer, lambda: pico_instance)
            self._metadata[PicoContainer] = ProviderMetadata(
                key=PicoContainer,
                provided_type=PicoContainer,
                concrete_class=PicoContainer,
                factory_class=None,
                factory_method=None,
                qualifiers=set(),
                primary=True,
                lazy=False,
                infra="component",
                pico_name="PicoContainer",
                override=True,
                scope=SCOPE_SINGLETON,
                dependencies=()
            )

        self._promote_scopes()
        self._rebuild_indexes()

        for _, selector, default_cls in sorted(on_missing, key=lambda x: -x[0]):
            key = selector
            if key in self._metadata or self._factory.has(key) or _can_be_selected_for(self._metadata, selector):
                continue

            deps = analyze_callable_dependencies(default_cls.__init__)
            provider = DeferredProvider(lambda pico, loc, c=default_cls, d=deps: pico.build_class(c, loc, d))
            qset = set(str(q) for q in getattr(default_cls, PICO_META, {}).get("qualifier", ()))
            sc = getattr(default_cls, PICO_META, {}).get("scope", SCOPE_SINGLETON)
            md = ProviderMetadata(key=key, provided_type=key if isinstance(key, type) else None, concrete_class=default_cls, factory_class=None, factory_method=None, qualifiers=qset, primary=True, lazy=bool(getattr(default_cls, PICO_META, {}).get("lazy", False)), infra=getattr(default_cls, PICO_INFRA, None), pico_name=getattr(default_cls, PICO_NAME, None), override=True, scope=sc, dependencies=deps)

            self._bind_if_absent(key, provider)
            self._metadata[key] = md
            if isinstance(provider, DeferredProvider):
                self._deferred.append(provider)

        self._rebuild_indexes()

        final_locator = ComponentLocator(self._metadata, self._indexes)
        validator = DependencyValidator(self._metadata, self._factory, final_locator)
        validator.validate_bindings()



import pytest
import sys
import json
import types
from dataclasses import dataclass
from typing import List, Dict, Union, Annotated, Optional

from pico_ioc import (
    init,
    configured,
    component,
    configuration,
    DictSource,
    JsonTreeSource,
    YamlTreeSource,
    Discriminator,
    ComponentCreationError,
    ConfigurationError,
)

@dataclass
class DBSettings:
    host: str
    port: int
    user: Optional[str] = "default"

@dataclass
class RedisSettings:
    url: str

@dataclass
class CacheSettings:
    ttl: int
    redis: RedisSettings

@dataclass
class FeatureConfig:
    features: List[str]

@dataclass
class User:
    name: str

@dataclass
class UserList:
    users: List[User]

@dataclass
class HeadersConfig:
    headers: Dict[str, str]

class SimpleService:
    def __init__(self, endpoint: str, timeout: int = 10):
        self.endpoint = endpoint
        self.timeout = timeout

@dataclass
class Cat:
    name: str
    lives: int = 9

@dataclass
class Dog:
    name: str
    breed: str

@dataclass
class PetOwnerDefault:
    pet: Union[Cat, Dog]

@dataclass
class PetOwnerCustom:
    pet: Annotated[Union[Cat, Dog], Discriminator("animal_type")]

@pytest.fixture
def temp_json_file(tmp_path):
    file_path = tmp_path / "config.json"
    def _create(data: dict):
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(data, f)
        return str(file_path)
    return _create

@pytest.fixture
def temp_yaml_file(tmp_path):
    file_path = tmp_path / "config.yaml"
    def _create(data: str):
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(data)
        return str(file_path)
    return _create

def test_configured_basic_dataclass():
    @configured(target=DBSettings, prefix="db")
    class ConfiguredDBSettings:
        pass
    
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettings", ConfiguredDBSettings)

    config_data = {"db": {"host": "localhost", "port": 5432}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    settings = container.get(DBSettings)
    
    assert isinstance(settings, DBSettings)
    assert settings.host == "localhost"
    assert settings.port == 5432
    assert settings.user == "default"

def test_configured_nested_dataclass():
    @configured(target=CacheSettings, prefix="cache")
    class ConfiguredCacheSettings:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredCacheSettings", ConfiguredCacheSettings)

    config_data = {"cache": {"ttl": 3600, "redis": {"url": "redis://host:6379"}}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    settings = container.get(CacheSettings)
    
    assert isinstance(settings, CacheSettings)
    assert settings.ttl == 3600
    assert isinstance(settings.redis, RedisSettings)
    assert settings.redis.url == "redis://host:6379"

def test_configured_list_of_primitives():
    @configured(target=FeatureConfig, prefix="app")
    class ConfiguredFeatureConfig:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredFeatureConfig", ConfiguredFeatureConfig)
    
    config_data = {"app": {"features": ["foo", "bar", "baz"]}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    config = container.get(FeatureConfig)
    
    assert isinstance(config, FeatureConfig)
    assert config.features == ["foo", "bar", "baz"]

def test_configured_list_of_dataclasses():
    @configured(target=UserList, prefix="user_config")
    class ConfiguredUserList:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredUserList", ConfiguredUserList)

    config_data = {"user_config": {"users": [{"name": "Alice"}, {"name": "Bob"}]}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    config = container.get(UserList)
    
    assert isinstance(config, UserList)
    assert len(config.users) == 2
    assert isinstance(config.users[0], User)
    assert config.users[0].name == "Alice"
    assert config.users[1].name == "Bob"

def test_configured_dict():
    @configured(target=HeadersConfig, prefix="api")
    class ConfiguredHeadersConfig:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredHeadersConfig", ConfiguredHeadersConfig)
    
    config_data = {"api": {"headers": {"X-Auth": "token123", "User-Agent": "pico"}}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    config = container.get(HeadersConfig)
    
    assert isinstance(config, HeadersConfig)
    assert config.headers == {"X-Auth": "token123", "User-Agent": "pico"}

def test_configured_plain_class_init():
    @configured(target=SimpleService, prefix="service")
    class ConfiguredSimpleService:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredSimpleService", ConfiguredSimpleService)

    config_data = {"service": {"endpoint": "http://api.com", "timeout": 5}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    service = container.get(SimpleService)
    
    assert isinstance(service, SimpleService)
    assert service.endpoint == "http://api.com"
    assert service.timeout == 5

def test_configured_json_tree_source(temp_json_file):
    @configured(target=DBSettings, prefix="db_file")
    class ConfiguredDBSettingsFile:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettingsFile", ConfiguredDBSettingsFile)
    
    config_path = temp_json_file({"db_file": {"host": "file.host", "port": 1234, "user": "file_user"}})
    ctx = configuration(JsonTreeSource(config_path))
    container = init(modules=[mod], config=ctx)
    settings = container.get(DBSettings)
    
    assert isinstance(settings, DBSettings)
    assert settings.host == "file.host"
    assert settings.port == 1234
    assert settings.user == "file_user"

def test_configured_yaml_tree_source(temp_yaml_file):
    yaml = pytest.importorskip("yaml")
    
    @configured(target=DBSettings, prefix="db_yaml")
    class ConfiguredDBSettingsYaml:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettingsYaml", ConfiguredDBSettingsYaml)
    
    config_content = """
db_yaml:
  host: yaml.host
  port: 5678
"""
    config_path = temp_yaml_file(config_content)
    ctx = configuration(YamlTreeSource(config_path))
    container = init(modules=[mod], config=ctx)
    settings = container.get(DBSettings)
    
    assert isinstance(settings, DBSettings)
    assert settings.host == "yaml.host"
    assert settings.port == 5678

def test_configured_env_interpolation(monkeypatch):
    @configured(target=DBSettings, prefix="db_env")
    class ConfiguredDBSettingsEnv:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettingsEnv", ConfiguredDBSettingsEnv)

    monkeypatch.setenv("DB_HOST_ENV", "env.host")
    config_data = {"db_env": {"host": "${ENV:DB_HOST_ENV}", "port": 9999}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    settings = container.get(DBSettings)
    
    assert settings.host == "env.host"
    assert settings.port == 9999

def test_configured_ref_interpolation():
    @configured(target=DBSettings, prefix="db_ref")
    class ConfiguredDBSettingsRef:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettingsRef", ConfiguredDBSettingsRef)

    config_data = {"defaults": {"user": "ref_user"}, "db_ref": {"host": "ref.host", "port": 1111, "user": "${ref:defaults.user}"}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    settings = container.get(DBSettings)
    
    assert settings.host == "ref.host"
    assert settings.port == 1111
    assert settings.user == "ref_user"

def test_configured_error_on_missing_prefix():
    @configured(target=DBSettings, prefix="missing.prefix")
    class ConfiguredDBSettingsMissing:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettingsMissing", ConfiguredDBSettingsMissing)

    ctx = configuration(DictSource({}))
    
    with pytest.raises(ComponentCreationError) as e:
        init(modules=[mod], config=ctx)
        
    assert isinstance(e.value.cause, ConfigurationError)
    assert "Missing config prefix: missing.prefix" in str(e.value.cause)

def test_configured_union_default_discriminator():
    @configured(target=PetOwnerDefault, prefix="pet_owner_default")
    class ConfiguredPetOwnerDefault:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredPetOwnerDefault", ConfiguredPetOwnerDefault)

    config_data = {"pet_owner_default": {"pet": {"$type": "Cat", "name": "Fluffy"}}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    owner = container.get(PetOwnerDefault)
    
    assert isinstance(owner.pet, Cat)
    assert owner.pet.name == "Fluffy"

def test_configured_union_custom_discriminator():
    @configured(target=PetOwnerCustom, prefix="pet_owner_custom")
    class ConfiguredPetOwnerCustom:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredPetOwnerCustom", ConfiguredPetOwnerCustom)

    config_data = {"pet_owner_custom": {"pet": {"animal_type": "Dog", "name": "Buddy", "breed": "Labrador"}}}
    ctx = configuration(DictSource(config_data))
    container = init(modules=[mod], config=ctx)
    owner = container.get(PetOwnerCustom)
    
    assert isinstance(owner.pet, Dog)
    assert owner.pet.name == "Buddy"
    assert owner.pet.breed == "Labrador"

def test_configured_selects_existing_and_longest_prefix():
    @configured(target=DBSettings, prefix="db")
    class CfgA:
        pass
    @configured(target=DBSettings, prefix="db_prod")
    class CfgB:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "CfgA", CfgA)
    setattr(mod, "CfgB", CfgB)
    
    ctx = configuration(DictSource({"db_prod": {"host": "h", "port": 1}}))
    container = init([mod], config=ctx)
    s = container.get(DBSettings)
    
    assert s.host == "h"

def test_configured_sets_config_metadata():
    @configured(target=DBSettings, prefix="db")
    class ConfiguredDBSettings:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredDBSettings", ConfiguredDBSettings)

    data = {"db": {"host": "x", "port": 1}}
    ctx = configuration(DictSource(data))
    c = init([mod], config=ctx)
    s = c.get(DBSettings)
    meta = getattr(s, "_pico_meta")
    
    assert meta["config_prefix"] == "db"
    assert "config_hash" in meta
    assert tuple(meta["config_source_order"])

def test_union_wrong_discriminator_raises_clear_error():
    @configured(target=PetOwnerDefault, prefix="pet_owner_default")
    class ConfiguredPetOwnerDefault:
        pass

    mod = types.ModuleType("test_mod")
    setattr(mod, "ConfiguredPetOwnerDefault", ConfiguredPetOwnerDefault)

    data = {"pet_owner_default": {"pet": {"$type": "Bird", "name": "Kiwi"}}}
    ctx = configuration(DictSource(data))
    
    with pytest.raises(ComponentCreationError) as e:
        init([mod], config=ctx)
        
    assert "Discriminator $type did not match" in str(e.value.cause)

def test_configured_typed_dict_and_type_errors():
    @dataclass
    class D:
        m: Dict[str, int]
    @configured(target=D, prefix="d")
    class C:
        pass
        
    mod = types.ModuleType("test_mod")
    setattr(mod, "D", D)
    setattr(mod, "C", C)
    
    ctx = configuration(DictSource({"d": {"m": {"a": "1"}}}))
    c = init([mod], config=ctx)
    assert c.get(D).m == {"a": 1}

def test_string_injection_resolves_by_pico_name_or_classname():
    @component(name="SpecialWidget")
    class W:
        pass
    @component
    class Needs:
        def __init__(self, w: "SpecialWidget"):
            self.w = w
            
    mod = types.ModuleType("test_mod")
    setattr(mod, "W", W)
    setattr(mod, "Needs", Needs)
    
    c = init([mod])
    assert isinstance(c.get(Needs).w, W)



"""
Additional tests to boost container.py coverage to 90%+.
Tests edge cases and less common code paths.
"""
import asyncio
import inspect
import tempfile
import os
import pytest
from typing import List, Dict, Any, Type
from unittest.mock import MagicMock, patch, AsyncMock

from pico_ioc import init, component, factory, provides, configure, cleanup
from pico_ioc.container import (
    PicoContainer,
    _get_signature_safe,
    _build_resolution_graph,
    _needs_async_configure,
    _iter_configure_methods,
)
from pico_ioc.aop import UnifiedComponentProxy, intercepted_by, MethodInterceptor, MethodCtx
from pico_ioc.scope import ScopedCaches, ScopeManager
from pico_ioc.factory import ComponentFactory, ProviderMetadata
from pico_ioc.locator import ComponentLocator
from pico_ioc.exceptions import AsyncResolutionError, ProviderNotFoundError
from pico_ioc.constants import PICO_META


# ============================================================
# Helper function tests - don't need container
# ============================================================

class TestGetSignatureSafeEdgeCases:
    """Test _get_signature_safe edge cases."""

    def test_signature_with_wrapped_attribute(self):
        """Function with __wrapped__ falls back to wrapped signature."""
        def original(x: int, y: str) -> bool:
            return True

        class BadCallable:
            def __call__(self):
                pass

        bad = BadCallable()
        bad.__wrapped__ = original

        sig = _get_signature_safe(bad)
        assert 'x' in sig.parameters
        assert 'y' in sig.parameters

    def test_signature_normal_function(self):
        """Normal function returns signature directly."""
        def normal_func(a: int, b: str = "default") -> None:
            pass

        sig = _get_signature_safe(normal_func)
        assert 'a' in sig.parameters
        assert 'b' in sig.parameters

    def test_signature_raises_when_no_wrapped(self):
        """_get_signature_safe raises when signature fails and no __wrapped__."""
        class BadObj:
            """Object that can't be introspected."""
            pass

        bad = BadObj()
        # Should raise since no __wrapped__ and signature fails
        with pytest.raises((ValueError, TypeError)):
            _get_signature_safe(bad)


class TestNeedsAsyncConfigure:
    """Test _needs_async_configure helper."""

    def test_needs_async_configure_detects_async(self):
        """_needs_async_configure returns True for async configure methods."""
        class Service:
            async def setup(self):
                pass

        Service.setup._pico_meta = {"configure": True}

        instance = Service()
        assert _needs_async_configure(instance) is True

    def test_needs_async_configure_false_for_sync(self):
        """_needs_async_configure returns False for sync configure."""
        class Service:
            def setup(self):
                pass

        Service.setup._pico_meta = {"configure": True}

        instance = Service()
        assert _needs_async_configure(instance) is False

    def test_needs_async_configure_false_no_configure(self):
        """Returns False when no configure methods exist."""
        class Service:
            def regular(self):
                pass

        instance = Service()
        assert _needs_async_configure(instance) is False


class TestIterConfigureMethods:
    """Test _iter_configure_methods helper."""

    def test_iter_configure_methods_yields_all(self):
        """_iter_configure_methods yields all configure methods."""
        class Service:
            def setup1(self):
                pass
            def setup2(self):
                pass
            def regular(self):
                pass

        Service.setup1._pico_meta = {"configure": True}
        Service.setup2._pico_meta = {"configure": True}

        instance = Service()
        methods = list(_iter_configure_methods(instance))

        assert len(methods) == 2

    def test_iter_configure_methods_empty(self):
        """_iter_configure_methods yields nothing for no configure methods."""
        class Service:
            def regular(self):
                pass

        instance = Service()
        methods = list(_iter_configure_methods(instance))

        assert len(methods) == 0


class TestBuildResolutionGraph:
    """Test _build_resolution_graph."""

    def test_build_resolution_graph_none_locator(self):
        """Returns empty dict for None locator."""
        result = _build_resolution_graph(None)
        assert result == {}

    def test_build_resolution_graph_with_mock_locator(self):
        """_build_resolution_graph works with mock locator."""
        mock_md = MagicMock()
        mock_md.provided_type = str
        mock_md.concrete_class = str
        mock_md.primary = True

        mock_locator = MagicMock()
        mock_locator._metadata = {str: mock_md}
        mock_locator.dependency_keys_for_static.return_value = []
        mock_locator.find_key_by_name.return_value = None

        result = _build_resolution_graph(mock_locator)
        assert isinstance(result, dict)

    def test_build_resolution_graph_maps_dependencies(self):
        """_build_resolution_graph maps dependencies correctly."""
        mock_md = MagicMock()
        mock_md.provided_type = int
        mock_md.concrete_class = int
        mock_md.primary = True

        mock_locator = MagicMock()
        mock_locator._metadata = {int: mock_md}
        mock_locator.dependency_keys_for_static.return_value = ["named_dep"]
        mock_locator.find_key_by_name.return_value = str

        result = _build_resolution_graph(mock_locator)
        assert int in result
        assert str in result[int]


class TestContainerInit:
    """Test container initialization."""

    def test_container_id_generation_unique(self):
        """Container generates unique IDs."""
        id1 = PicoContainer._generate_container_id()
        id2 = PicoContainer._generate_container_id()

        assert id1 != id2
        assert id1.startswith('c')

    def test_container_custom_id(self):
        """Container accepts custom ID."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(
            factory_mock, caches, scopes,
            container_id="my-custom-id"
        )

        assert container.container_id == "my-custom-id"
        container.shutdown()

    def test_container_with_profiles(self):
        """Container stores profiles in context."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(
            factory_mock, caches, scopes,
            profiles=("dev", "test")
        )

        assert container.context.profiles == ("dev", "test")
        container.shutdown()


class TestContainerContext:
    """Test container context management."""

    def test_get_current_returns_none(self):
        """get_current returns None when no container active."""
        PicoContainer._container_id_var.set(None)
        assert PicoContainer.get_current() is None

    def test_get_current_id_returns_none(self):
        """get_current_id returns None when no container active."""
        PicoContainer._container_id_var.set(None)
        assert PicoContainer.get_current_id() is None

    def test_all_containers_returns_dict(self):
        """all_containers returns a dict."""
        result = PicoContainer.all_containers()
        assert isinstance(result, dict)


class TestContainerRegistryManagement:
    """Test container registry management."""

    def test_shutdown_removes_from_registry(self):
        """shutdown() removes container from registry."""
        container = init(modules=[])
        cid = container.container_id

        assert cid in PicoContainer._container_registry

        container.shutdown()

        assert cid not in PicoContainer._container_registry

    @pytest.mark.asyncio
    async def test_ashutdown_removes_from_registry(self):
        """ashutdown() removes container from registry."""
        container = init(modules=[])
        cid = container.container_id

        assert cid in PicoContainer._container_registry

        await container.ashutdown()

        assert cid not in PicoContainer._container_registry


class TestScopeContextManager:
    """Test container.scope() context manager."""

    def test_scope_context_manager_activates_deactivates(self):
        """scope() context manager properly activates/deactivates scope."""
        container = init(modules=[])
        try:
            assert container.scopes.get_id("request") is None

            with container.scope("request", "test-id"):
                assert container.scopes.get_id("request") == "test-id"

            assert container.scopes.get_id("request") is None
        finally:
            container.shutdown()


class TestExportGraphWithoutLocator:
    """Test export_graph error handling."""

    def test_export_graph_raises_without_locator(self):
        """export_graph raises if no locator attached."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(factory_mock, caches, scopes)

        try:
            with pytest.raises(RuntimeError, match="No locator attached"):
                container.export_graph("/tmp/test.dot")
        finally:
            container.shutdown()


class TestCanonicalKeyUnknown:
    """Test _canonical_key with unknown keys."""

    def test_canonical_key_unknown_returns_same(self):
        """_canonical_key returns unknown key unchanged."""
        container = init(modules=[])
        try:
            key = container._canonical_key("unknown_string_key")
            assert key == "unknown_string_key"
        finally:
            container.shutdown()


class TestContainerActivateDeactivate:
    """Test container activate/deactivate."""

    def test_activate_sets_context_var(self):
        """activate() sets the container ID context var."""
        container = init(modules=[])
        try:
            token = container.activate()
            assert PicoContainer._container_id_var.get() == container.container_id
            container.deactivate(token)
        finally:
            container.shutdown()

    def test_as_current_context_manager(self):
        """as_current() context manager works correctly."""
        container = init(modules=[])
        try:
            with container.as_current() as c:
                assert c is container
                assert PicoContainer.get_current() is container
        finally:
            container.shutdown()


class TestContainerStats:
    """Test container stats."""

    def test_stats_returns_all_fields(self):
        """stats() returns all expected fields."""
        container = init(modules=[])
        try:
            stats = container.stats()

            assert "container_id" in stats
            assert "profiles" in stats
            assert "uptime_seconds" in stats
            assert "total_resolves" in stats
            assert "cache_hits" in stats
            assert "cache_hit_rate" in stats
            assert "registered_components" in stats
        finally:
            container.shutdown()

    def test_stats_cache_hit_rate_zero_when_empty(self):
        """cache_hit_rate is 0.0 when no resolves."""
        container = init(modules=[])
        try:
            stats = container.stats()
            # No resolves means 0 hit rate
            assert stats["cache_hit_rate"] == 0.0
        finally:
            container.shutdown()


class TestContainerInfo:
    """Test container info logging."""

    def test_info_logs_message(self, caplog):
        """info() logs a message."""
        import logging

        container = init(modules=[])
        try:
            with caplog.at_level(logging.INFO, logger="pico_ioc"):
                container.info("Test info message")

            assert any("Test info message" in r.message for r in caplog.records)
        finally:
            container.shutdown()


class TestCleanupAllAsync:
    """Test cleanup_all_async."""

    @pytest.mark.asyncio
    async def test_cleanup_all_async_runs(self):
        """cleanup_all_async executes without error."""
        container = init(modules=[])
        try:
            await container.cleanup_all_async()
        finally:
            PicoContainer._container_registry.pop(container.container_id, None)


class TestHasMethod:
    """Test has() method."""

    def test_has_returns_false_for_unknown(self):
        """has() returns False for unknown keys."""
        container = init(modules=[])
        try:
            class UnknownClass:
                pass

            assert container.has(UnknownClass) is False
            assert container.has("unknown_key") is False
        finally:
            container.shutdown()


class TestAttachLocator:
    """Test attach_locator."""

    def test_attach_locator_sets_locator(self):
        """attach_locator() sets the _locator attribute."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(factory_mock, caches, scopes)

        assert container._locator is None

        mock_locator = MagicMock(spec=ComponentLocator)
        # Mock the _metadata attribute that shutdown() iterates
        mock_locator._metadata = {}
        container.attach_locator(mock_locator)

        assert container._locator is mock_locator
        container.shutdown()


class TestActivateDeactivateScope:
    """Test scope activation/deactivation."""

    def test_activate_scope_delegates_to_scopes(self):
        """activate_scope() delegates to scopes manager."""
        container = init(modules=[])
        try:
            token = container.activate_scope("request", "req-123")
            assert container.scopes.get_id("request") == "req-123"

            container.deactivate_scope("request", token)
            assert container.scopes.get_id("request") is None
        finally:
            container.shutdown()


class TestBuildResolutionGraphMethod:
    """Test build_resolution_graph method."""

    def test_build_resolution_graph_with_no_locator(self):
        """build_resolution_graph returns empty for no locator."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(factory_mock, caches, scopes)

        graph = container.build_resolution_graph()
        assert graph == {}

        container.shutdown()


# ============================================================
# Additional tests for uncovered lines - using mocks
# ============================================================

class TestBuildResolutionGraphIssubclassException:
    """Test _build_resolution_graph handles issubclass exceptions."""

    def test_handles_non_type_provided_type(self):
        """_build_resolution_graph handles non-type provided_type."""
        mock_md = MagicMock()
        mock_md.provided_type = "not_a_type"  # String, not type
        mock_md.concrete_class = "also_not_type"
        mock_md.primary = True

        mock_locator = MagicMock()
        mock_locator._metadata = {"key": mock_md}
        mock_locator.dependency_keys_for_static.return_value = [int]
        mock_locator.find_key_by_name.return_value = None

        # Should not raise - just continues
        result = _build_resolution_graph(mock_locator)
        assert isinstance(result, dict)


class TestCanonicalKeyWithTypeNotFound:
    """Test _canonical_key with type that has no subclass match."""

    def test_canonical_key_unknown_type_returns_same(self):
        """_canonical_key returns unknown type unchanged."""
        container = init(modules=[])
        try:
            class SomeRandomClass:
                pass

            key = container._canonical_key(SomeRandomClass)
            assert key is SomeRandomClass
        finally:
            container.shutdown()


class TestExportGraphWritesFile:
    """Test export_graph file writing."""

    def test_export_graph_writes_dot_file(self):
        """export_graph writes valid DOT file."""
        container = init(modules=[])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path)

            with open(path, 'r') as f:
                content = f.read()

            assert 'digraph Pico' in content
            os.unlink(path)
        finally:
            container.shutdown()

    def test_export_graph_with_title(self):
        """export_graph includes title when specified."""
        container = init(modules=[])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path, title="My App")

            with open(path, 'r') as f:
                content = f.read()

            assert 'My App' in content
            os.unlink(path)
        finally:
            container.shutdown()


class TestContainerRepr:
    """Test container __repr__."""

    def test_container_repr(self):
        """Container has a useful repr."""
        container = init(modules=[])
        try:
            r = repr(container)
            assert "PicoContainer" in r or "container" in r.lower()
        finally:
            container.shutdown()



from typing import Any, Callable, Dict, List, Optional, Tuple, Union, Set
from .factory import ComponentFactory, ProviderMetadata
from .locator import ComponentLocator
from .exceptions import InvalidBindingError
from .analysis import DependencyRequest

KeyT = Union[str, type]

def _fmt(k: KeyT) -> str:
    return getattr(k, '__name__', str(k))

def _skip_type(t: type) -> bool:
    if t in (str, int, float, bool, bytes):
        return True
    if t is Any:
        return True
    if getattr(t, "_is_protocol", False):
        return True
    return False

class DependencyValidator:
    def __init__(self, metadata: Dict[KeyT, ProviderMetadata], factory: ComponentFactory, locator: ComponentLocator):
        self._metadata = metadata
        self._factory = factory
        self._locator = locator

    def _find_md_for_type(self, t: type) -> Optional[ProviderMetadata]:
        cands: List[ProviderMetadata] = []
        for md in self._metadata.values():
            typ = md.provided_type or md.concrete_class
            if not isinstance(typ, type):
                continue
            try:
                if issubclass(typ, t):
                    cands.append(md)
            except TypeError:
                 pass
            except Exception:
                continue
        if not cands:
            if getattr(t, "_is_protocol", False):
                 for md in self._metadata.values():
                    typ = md.provided_type or md.concrete_class
                    if isinstance(typ, type) and ComponentLocator._implements_protocol(typ, t):
                        cands.append(md)

        if not cands:
            return None
        prim = [m for m in cands if m.primary]
        return prim[0] if prim else cands[0]


    def _find_md_for_name(self, name: str) -> Optional[KeyT]:
         return self._locator.find_key_by_name(name)


    def validate_bindings(self) -> None:
        errors: List[str] = []

        for k, md in self._metadata.items():
            if md.infra == "configuration":
                continue

            if not md.dependencies and md.infra not in ("configured", "component") and not md.override:
                 continue
            if md.infra == "component" and md.concrete_class and md.concrete_class.__init__ is object.__init__:
                 continue


            loc_name = f"component {_fmt(k)}"
            if md.factory_method:
                loc_name = f"factory method {md.factory_method}"

            for dep in md.dependencies:
                if dep.is_optional:
                    continue

                if dep.is_list:
                    if dep.qualifier:
                        if not self._locator.collect_by_type(dep.key, dep.qualifier) and isinstance(dep.key, type) and not _skip_type(dep.key):
                            errors.append(f"{_fmt(k)} ({loc_name}) expects List[{_fmt(dep.key)}] with qualifier '{dep.qualifier}' but no matching components exist")
                    continue

                dep_key = dep.key
                if isinstance(dep_key, str):
                    key_found_by_name = self._find_md_for_name(dep_key)
                    directly_bound = dep_key in self._metadata or self._factory.has(dep_key)
                    if not key_found_by_name and not directly_bound:
                        errors.append(f"{_fmt(k)} ({loc_name}) depends on string key '{dep_key}' which is not bound")
                    continue

                if isinstance(dep_key, type) and not _skip_type(dep_key):
                    dep_key_found = self._factory.has(dep_key) or dep_key in self._metadata
                    if not dep_key_found:
                        assignable_md = self._find_md_for_type(dep_key)
                        if assignable_md is None:
                            by_name_key = self._find_md_for_name(getattr(dep_key, "__name__", ""))
                            if by_name_key is None:
                                errors.append(f"{_fmt(k)} ({loc_name}) depends on {_fmt(dep_key)} which is not bound")
                    continue

        if errors:
            raise InvalidBindingError(errors)



import inspect
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, Set, Protocol
from .constants import PICO_INFRA, PICO_NAME, PICO_KEY, PICO_META, SCOPE_SINGLETON
from .factory import ProviderMetadata, DeferredProvider
from .decorators import get_return_type
from .config_registrar import ConfigurationManager
from .analysis import analyze_callable_dependencies, DependencyRequest

KeyT = Union[str, type]
Provider = Callable[[], Any]

class CustomScanner(Protocol):
    def should_scan(self, obj: Any) -> bool: ...
    def scan(self, obj: Any) -> Optional[Tuple[KeyT, Provider, ProviderMetadata]]: ...

class ComponentScanner:
    def __init__(self, profiles: Set[str], environ: Dict[str, str], config_manager: ConfigurationManager):
        self._profiles = profiles
        self._environ = environ
        self._config_manager = config_manager
        self._candidates: Dict[KeyT, List[Tuple[bool, Provider, ProviderMetadata]]] = {}
        self._on_missing: List[Tuple[int, KeyT, type]] = []
        self._deferred: List[DeferredProvider] = []
        self._provides_functions: Dict[KeyT, Callable[..., Any]] = {}
        self._custom_scanners: List[CustomScanner] = []

    def register_custom_scanner(self, scanner: CustomScanner) -> None:
        self._custom_scanners.append(scanner)

    def get_scan_results(self) -> Tuple[Dict[KeyT, List[Tuple[bool, Provider, ProviderMetadata]]], List[Tuple[int, KeyT, type]], List[DeferredProvider], Dict[KeyT, Callable[..., Any]]]:
        return self._candidates, self._on_missing, self._deferred, self._provides_functions

    def _queue(self, key: KeyT, provider: Provider, md: ProviderMetadata) -> None:
        lst = self._candidates.setdefault(key, [])
        lst.append((md.primary, provider, md))
        if isinstance(provider, DeferredProvider):
            self._deferred.append(provider)

    def _enabled_by_condition(self, obj: Any) -> bool:
        meta = getattr(obj, PICO_META, {})
        c = meta.get("conditional", None)
        if not c:
            return True

        p = set(c.get("profiles") or ())
        if p and not (p & self._profiles):
            return False

        req = c.get("require_env") or ()
        for k in req:
            if k not in self._environ or not self._environ.get(k):
                return False

        pred = c.get("predicate")
        if pred is None:
            return True

        try:
            ok = bool(pred())
        except Exception:
            return False
        if not ok:
            return False

        return True

    def _register_component_class(self, cls: type) -> None:
        if not self._enabled_by_condition(cls):
            return
        key = getattr(cls, PICO_KEY, cls)
        qset = set(str(q) for q in getattr(cls, PICO_META, {}).get("qualifier", ()))
        sc = getattr(cls, PICO_META, {}).get("scope", SCOPE_SINGLETON)
        deps = analyze_callable_dependencies(cls.__init__)
        provider = DeferredProvider(lambda pico, loc, c=cls, d=deps: pico.build_class(c, loc, d))
        md = ProviderMetadata(key=key, provided_type=cls, concrete_class=cls, factory_class=None, factory_method=None, qualifiers=qset, primary=bool(getattr(cls, PICO_META, {}).get("primary")), lazy=bool(getattr(cls, PICO_META, {}).get("lazy", False)), infra=getattr(cls, PICO_INFRA, None), pico_name=getattr(cls, PICO_NAME, None), scope=sc, dependencies=deps)
        self._queue(key, provider, md)

    def _register_factory_class(self, cls: type) -> None:
        if not self._enabled_by_condition(cls):
            return

        factory_deps: Optional[Tuple[DependencyRequest, ...]] = None
        has_instance_provides = False
        for name in dir(cls):
                 try:
                     raw = inspect.getattr_static(cls, name)
                     if inspect.isfunction(raw) and getattr(raw, PICO_INFRA, None) == "provides":
                         has_instance_provides = True
                         break
                 except Exception:
                     continue

        if has_instance_provides:
            factory_deps = analyze_callable_dependencies(cls.__init__)

        for name in dir(cls):
            try:
                raw = inspect.getattr_static(cls, name)
            except Exception:
                continue

            fn = None
            kind = None
            if isinstance(raw, staticmethod):
                fn = raw.__func__
                kind = "static"
            elif isinstance(raw, classmethod):
                fn = raw.__func__
                kind = "class"
            elif inspect.isfunction(raw):
                fn = raw
                kind = "instance"
            else:
                continue

            if getattr(fn, PICO_INFRA, None) != "provides":
                continue
            if not self._enabled_by_condition(fn):
                continue

            k = getattr(fn, PICO_KEY)
            deps = analyze_callable_dependencies(fn)

            if kind == "instance":
                if factory_deps is None: factory_deps = analyze_callable_dependencies(cls.__init__)
                provider = DeferredProvider(lambda pico, loc, fc=cls, mn=name, df=factory_deps, dm=deps: pico.build_method(getattr(pico.build_class(fc, loc, df), mn), loc, dm))
            else:
                provider = DeferredProvider(lambda pico, loc, f=fn, d=deps: pico.build_method(f, loc, d))

            rt = get_return_type(fn)
            qset = set(str(q) for q in getattr(fn, PICO_META, {}).get("qualifier", ()))
            sc = getattr(fn, PICO_META, {}).get("scope", getattr(cls, PICO_META, {}).get("scope", SCOPE_SINGLETON))
            md = ProviderMetadata(key=k, provided_type=rt if isinstance(rt, type) else (k if isinstance(k, type) else None), concrete_class=None, factory_class=cls, factory_method=name, qualifiers=qset, primary=bool(getattr(fn, PICO_META, {}).get("primary")), lazy=bool(getattr(fn, PICO_META, {}).get("lazy", False)), infra=getattr(cls, PICO_INFRA, None), pico_name=getattr(fn, PICO_NAME, None), scope=sc, dependencies=deps)
            self._queue(k, provider, md)

    def _register_provides_function(self, fn: Callable[..., Any]) -> None:
        if not self._enabled_by_condition(fn):
            return
        k = getattr(fn, PICO_KEY)
        deps = analyze_callable_dependencies(fn)
        provider = DeferredProvider(lambda pico, loc, f=fn, d=deps: pico.build_method(f, loc, d))
        rt = get_return_type(fn)
        qset = set(str(q) for q in getattr(fn, PICO_META, {}).get("qualifier", ()))
        sc = getattr(fn, PICO_META, {}).get("scope", SCOPE_SINGLETON)
        md = ProviderMetadata(key=k, provided_type=rt if isinstance(rt, type) else (k if isinstance(k, type) else None), concrete_class=None, factory_class=None, factory_method=getattr(fn, "__name__", None), qualifiers=qset, primary=bool(getattr(fn, PICO_META, {}).get("primary")), lazy=bool(getattr(fn, PICO_META, {}).get("lazy", False)), infra="provides", pico_name=getattr(fn, PICO_NAME, None), scope=sc, dependencies=deps)
        self._queue(k, provider, md)
        self._provides_functions[k] = fn

    def _try_custom_scanners(self, obj: Any) -> bool:
        for scanner in self._custom_scanners:
            if scanner.should_scan(obj):
                result = scanner.scan(obj)
                if result:
                    key, provider, md = result
                    self._queue(key, provider, md)
                    if isinstance(provider, DeferredProvider):
                        self._deferred.append(provider)
                    return True
        return False

    def scan_module(self, module: Any) -> None:
        for _, obj in inspect.getmembers(module):
            if self._try_custom_scanners(obj):
                continue

            if inspect.isclass(obj) or getattr(obj, "_is_protocol", False):
                if inspect.isclass(obj):
                    meta = getattr(obj, PICO_META, {})
                    
                    if "on_missing" in meta:
                        sel = meta["on_missing"]["selector"]
                        pr = int(meta["on_missing"].get("priority", 0))
                        self._on_missing.append((pr, sel, obj))
                        continue

                    infra = getattr(obj, PICO_INFRA, None)
                    if infra == "component":
                        self._register_component_class(obj)
                    elif infra == "factory":
                        self._register_factory_class(obj)
                    elif infra == "configured":
                        enabled = self._enabled_by_condition(obj)
                        reg_data = self._config_manager.register_configured_class(obj, enabled)
                        if reg_data:
                            self._queue(reg_data[0], reg_data[1], reg_data[2])

            elif inspect.isfunction(obj):
                if getattr(obj, PICO_INFRA, None) == "provides":
                    self._register_provides_function(obj)



import pytest
from typing import List, Union, Optional
from dataclasses import dataclass
from pico_ioc.config_registrar import _coerce, ConfigurationManager
from pico_ioc.config_builder import ContextConfig, FlatDictSource
from pico_ioc.exceptions import ConfigurationError
from pico_ioc.factory import ProviderMetadata

def test_coerce_edge_cases():
    assert _coerce(None, int) is None
    assert _coerce("1.5", float) == 1.5
    assert _coerce("true", bool) is True

def test_config_manager_flat_lookup_fail():
    mgr = ConfigurationManager(None)
    assert mgr._lookup_flat("NON_EXISTENT") is None

def test_register_configured_class_validation():
    mgr = ConfigurationManager(None)
    
    assert mgr.register_configured_class(int, enabled=False) is None
    assert mgr.register_configured_class(int, enabled=True) is None
    
    class BadTarget:
        _pico_meta = {"configured": {"target": "not_a_type"}}
        
    assert mgr.register_configured_class(BadTarget, enabled=True) is None

def test_build_flat_not_dataclass():
    mgr = ConfigurationManager(None)
    with pytest.raises(ConfigurationError, match="must be a dataclass"):
        mgr._build_flat_instance(int, prefix="")

def test_prefix_exists_checks():
    mgr = ConfigurationManager(None)
    md = ProviderMetadata(
        key="k", provided_type=int, concrete_class=int, 
        factory_class=None, factory_method=None, qualifiers=set(),
        primary=False, lazy=False, infra="component", pico_name="p", scope="s"
    )
    
    assert mgr.prefix_exists(md) is False
    
    md_bad = ProviderMetadata(
        key="k", provided_type=None, concrete_class=None,
        factory_class=None, factory_method=None, qualifiers=set(),
        primary=False, lazy=False, infra="configured", pico_name="p", scope="s"
    )
    assert mgr.prefix_exists(md_bad) is False



import pytest
from dataclasses import dataclass
from typing import Any, Optional, Tuple, Callable
from pico_ioc import init
from pico_ioc.factory import DeferredProvider, ProviderMetadata

class PlainService:
    def get_value(self) -> str:
        return "I am a plain service found by custom scanner"

class MagicMarkerScanner:
    def should_scan(self, obj: Any) -> bool:
        return isinstance(obj, type) and getattr(obj, "__name__", "") == "PlainService"

    def scan(self, obj: Any) -> Optional[Tuple[Any, Callable[[], Any], ProviderMetadata]]:
        if not self.should_scan(obj):
            return None
        
        def builder(pico, loc):
            return obj()

        provider = DeferredProvider(builder)
        
        metadata = ProviderMetadata(
            key=obj,
            provided_type=obj,
            concrete_class=obj,
            factory_class=None,
            factory_method=None,
            qualifiers=set(),
            primary=True,
            lazy=False,
            infra="custom_magic",
            pico_name="plain_service",
            scope="singleton",
            dependencies=()
        )
        
        return (obj, provider, metadata)

def test_custom_scanner_registration_and_resolution():
    scanner = MagicMarkerScanner()
    
    container = init(
        modules=[__name__],
        custom_scanners=[scanner]
    )
    
    assert container.has(PlainService)
    
    instance = container.get(PlainService)
    assert isinstance(instance, PlainService)
    assert instance.get_value() == "I am a plain service found by custom scanner"

def test_custom_scanner_is_ignored_if_not_registered():
    container = init(
        modules=[__name__]
    )
    
    assert not container.has(PlainService)
    
    with pytest.raises(Exception):
        container.get(PlainService)



"""
Extended tests for container.py to increase coverage.
"""
import asyncio
import inspect
import os
import tempfile
import pytest
from unittest.mock import MagicMock, patch, AsyncMock

from pico_ioc import init, component, factory, provides, configure, cleanup
from pico_ioc.container import (
    PicoContainer,
    _normalize_callable,
    _get_signature_safe,
    _needs_async_configure,
    _iter_configure_methods,
    _build_resolution_graph
)
from pico_ioc.constants import PICO_META
from pico_ioc.exceptions import (
    ComponentCreationError,
    ProviderNotFoundError,
    AsyncResolutionError,
    ConfigurationError
)
from pico_ioc.factory import ComponentFactory
from pico_ioc.scope import ScopedCaches, ScopeManager
from pico_ioc.aop import ContainerObserver, health


# Define test components at module level for proper scanning
@component(name="named_service")
class NamedService:
    pass


@component()
class RegisteredService:
    pass


@component()
class ExportDatabase:
    pass


@component()
class ExportRepository:
    def __init__(self, db: ExportDatabase):
        self.db = db


@component()
class CleanableService:
    cleanup_called = False

    @cleanup
    def close(self):
        CleanableService.cleanup_called = True


@component()
class AsyncCleanableService:
    cleanup_called = False

    @cleanup
    async def close(self):
        await asyncio.sleep(0.01)
        AsyncCleanableService.cleanup_called = True


@component()
class StatsTestService:
    pass


@component()
class TrackedService:
    pass


@component()
class CachedService:
    pass


@component()
class ObservedService:
    pass


@component()
class CacheObservedService:
    pass


@component()
class HealthyService:
    @health
    def is_healthy(self) -> bool:
        return True


@component()
class UnhealthyService:
    @health
    def is_healthy(self) -> bool:
        raise RuntimeError("Health check failed")


class TestContainerHelperFunctions:
    """Test helper functions in container module."""

    def test_normalize_callable_with_func_attr(self):
        """_normalize_callable returns __func__ if present."""
        class MyClass:
            def method(self):
                pass

        bound = MyClass().method
        result = _normalize_callable(bound)
        assert result == bound.__func__

    def test_normalize_callable_without_func_attr(self):
        """_normalize_callable returns obj if no __func__."""
        def plain_func():
            pass

        result = _normalize_callable(plain_func)
        assert result is plain_func

    def test_get_signature_safe_normal(self):
        """_get_signature_safe works with normal functions."""
        def my_func(a: int, b: str) -> bool:
            pass

        sig = _get_signature_safe(my_func)
        assert 'a' in sig.parameters
        assert 'b' in sig.parameters

    def test_get_signature_safe_wrapped(self):
        """_get_signature_safe falls back to __wrapped__."""
        def original(a: int) -> int:
            return a

        def wrapper(*args, **kwargs):
            return original(*args, **kwargs)

        wrapper.__wrapped__ = original

        sig = _get_signature_safe(wrapper)
        assert 'a' in sig.parameters

    def test_needs_async_configure_true(self):
        """_needs_async_configure detects async configure methods."""
        class Service:
            async def setup(self):
                pass

        Service.setup._pico_meta = {"configure": True}

        instance = Service()
        assert _needs_async_configure(instance) is True

    def test_needs_async_configure_false_sync(self):
        """_needs_async_configure returns False for sync configure."""
        class Service:
            def setup(self):
                pass

        Service.setup._pico_meta = {"configure": True}

        instance = Service()
        assert _needs_async_configure(instance) is False

    def test_needs_async_configure_false_no_configure(self):
        """_needs_async_configure returns False if no configure methods."""
        class Service:
            def regular_method(self):
                pass

        instance = Service()
        assert _needs_async_configure(instance) is False

    def test_iter_configure_methods(self):
        """_iter_configure_methods yields configure methods."""
        class Service:
            def setup(self):
                pass
            def teardown(self):
                pass
            def normal(self):
                pass

        Service.setup._pico_meta = {"configure": True}
        Service.teardown._pico_meta = {"configure": True}

        instance = Service()
        methods = list(_iter_configure_methods(instance))

        assert len(methods) == 2

    def test_build_resolution_graph_empty_locator(self):
        """_build_resolution_graph returns empty dict for None locator."""
        assert _build_resolution_graph(None) == {}


class TestContainerInit:
    """Test container initialization."""

    def test_container_id_generation(self):
        """Container generates unique IDs."""
        id1 = PicoContainer._generate_container_id()
        id2 = PicoContainer._generate_container_id()

        assert id1 != id2
        assert id1.startswith('c')
        assert id2.startswith('c')

    def test_container_custom_id(self):
        """Container accepts custom ID."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(
            factory_mock, caches, scopes,
            container_id="custom-123"
        )

        assert container.container_id == "custom-123"
        container.shutdown()

    def test_container_with_profiles(self):
        """Container stores profiles in context."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(
            factory_mock, caches, scopes,
            profiles=("dev", "test")
        )

        assert container.context.profiles == ("dev", "test")
        container.shutdown()

    def test_container_with_observers(self):
        """Container accepts observers."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        observer = MagicMock(spec=ContainerObserver)

        container = PicoContainer(
            factory_mock, caches, scopes,
            observers=[observer]
        )

        assert observer in container._observers
        container.shutdown()


class TestContainerContext:
    """Test container context management."""

    def test_get_current_returns_none_initially(self):
        """get_current returns None when no container active."""
        PicoContainer._container_id_var.set(None)

        assert PicoContainer.get_current() is None

    def test_get_current_id_returns_none_initially(self):
        """get_current_id returns None when no container active."""
        PicoContainer._container_id_var.set(None)

        assert PicoContainer.get_current_id() is None

    def test_all_containers_returns_copy(self):
        """all_containers returns a copy of registry."""
        result = PicoContainer.all_containers()

        assert isinstance(result, dict)


class TestContainerResolution:
    """Test component resolution."""

    def test_canonical_key_with_string_name(self):
        """_canonical_key maps string names to registered keys if found."""
        container = init(modules=[__name__])
        try:
            # When a pico_name matches, the key is mapped to the class
            # Note: This depends on how the locator finds the name
            key = container._canonical_key("named_service")
            # If not found as a type, returns the string as-is
            assert key == "named_service" or key is NamedService
        finally:
            container.shutdown()

    def test_canonical_key_with_unknown_returns_same(self):
        """_canonical_key returns unknown keys unchanged."""
        container = init(modules=[__name__])
        try:
            key = container._canonical_key("unknown_key")
            assert key == "unknown_key"
        finally:
            container.shutdown()

    def test_has_returns_true_for_registered(self):
        """has() returns True for registered components."""
        container = init(modules=[__name__])
        try:
            assert container.has(RegisteredService) is True
        finally:
            container.shutdown()

    def test_has_returns_false_for_unregistered(self):
        """has() returns False for unregistered types."""
        container = init(modules=[])
        try:
            class NotRegistered:
                pass

            assert container.has(NotRegistered) is False
        finally:
            container.shutdown()


class TestContainerExportGraph:
    """Test dependency graph export functionality."""

    def test_export_graph_to_file(self):
        """export_graph writes DOT file to disk."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path)

            with open(path, 'r') as f:
                content = f.read()

            assert 'digraph Pico' in content
            assert 'ExportDatabase' in content
            assert 'ExportRepository' in content
            assert '->' in content

            os.unlink(path)
        finally:
            container.shutdown()

    def test_export_graph_with_title(self):
        """export_graph includes title when specified."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path, title="My Application")

            with open(path, 'r') as f:
                content = f.read()

            assert 'My Application' in content
            assert 'labelloc="t"' in content

            os.unlink(path)
        finally:
            container.shutdown()

    def test_export_graph_with_qualifiers(self):
        """export_graph includes qualifiers when enabled."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path, include_qualifiers=True)

            with open(path, 'r') as f:
                content = f.read()

            assert 'digraph Pico' in content

            os.unlink(path)
        finally:
            container.shutdown()

    def test_export_graph_rankdir(self):
        """export_graph respects rankdir parameter."""
        container = init(modules=[__name__])
        try:
            with tempfile.NamedTemporaryFile(mode='w', suffix='.dot', delete=False) as f:
                path = f.name

            container.export_graph(path, rankdir="TB")

            with open(path, 'r') as f:
                content = f.read()

            assert 'rankdir="TB"' in content

            os.unlink(path)
        finally:
            container.shutdown()

    def test_export_graph_raises_without_locator(self):
        """export_graph raises error if no locator attached."""
        factory_mock = MagicMock(spec=ComponentFactory)
        caches = ScopedCaches()
        scopes = ScopeManager()

        container = PicoContainer(factory_mock, caches, scopes)

        try:
            with pytest.raises(RuntimeError, match="No locator attached"):
                container.export_graph("/tmp/test.dot")
        finally:
            container.shutdown()


class TestContainerCleanup:
    """Test cleanup functionality."""

    def test_cleanup_all_calls_cleanup_methods(self):
        """cleanup_all invokes @cleanup decorated methods."""
        CleanableService.cleanup_called = False

        container = init(modules=[__name__])
        container.get(CleanableService)

        container.cleanup_all()

        assert CleanableService.cleanup_called is True
        container.shutdown()

    @pytest.mark.asyncio
    async def test_cleanup_all_async_awaits_cleanup(self):
        """cleanup_all_async awaits async cleanup methods."""
        AsyncCleanableService.cleanup_called = False

        container = init(modules=[__name__])
        await container.aget(AsyncCleanableService)

        await container.cleanup_all_async()

        assert AsyncCleanableService.cleanup_called is True


class TestContainerStats:
    """Test container statistics."""

    def test_stats_returns_expected_fields(self):
        """stats() returns all expected fields."""
        container = init(modules=[__name__])
        try:
            stats = container.stats()

            assert "container_id" in stats
            assert "profiles" in stats
            assert "uptime_seconds" in stats
            assert "total_resolves" in stats
            assert "cache_hits" in stats
            assert "cache_hit_rate" in stats
            assert "registered_components" in stats
        finally:
            container.shutdown()

    def test_stats_tracks_resolves(self):
        """stats() tracks resolution count."""
        container = init(modules=[__name__])
        try:
            # Get initial count
            initial_resolves = container.context.resolve_count

            # Force a new resolution (TrackedService may already be resolved)
            # Use a scoped resolution to ensure a new resolve
            with container.scope("request", "test-req-123"):
                pass  # Just activate a scope

            # After any resolution, total_resolves should be >= 0
            stats = container.stats()
            assert stats["total_resolves"] >= 0
            assert "total_resolves" in stats
        finally:
            container.shutdown()

    def test_stats_tracks_cache_hits(self):
        """stats() tracks cache hit count."""
        container = init(modules=[__name__])
        try:
            # First call - miss
            container.get(CachedService)
            stats1 = container.stats()
            hits1 = stats1["cache_hits"]

            # Second call - hit
            container.get(CachedService)
            stats2 = container.stats()
            hits2 = stats2["cache_hits"]

            assert hits2 > hits1
        finally:
            container.shutdown()


class TestContainerObservers:
    """Test observer notifications."""

    def test_observers_notified_on_resolve(self):
        """Observers receive on_resolve notification."""
        observer = MagicMock(spec=ContainerObserver)

        container = init(modules=[__name__], observers=[observer])
        try:
            container.get(ObservedService)

            observer.on_resolve.assert_called()
        finally:
            container.shutdown()

    def test_observers_notified_on_cache_hit(self):
        """Observers receive on_cache_hit notification."""
        observer = MagicMock(spec=ContainerObserver)

        container = init(modules=[__name__], observers=[observer])
        try:
            # First call - resolve
            container.get(CacheObservedService)
            # Second call - cache hit
            container.get(CacheObservedService)

            observer.on_cache_hit.assert_called()
        finally:
            container.shutdown()


class TestContainerHealthCheck:
    """Test health check functionality."""

    def test_health_check_runs_health_methods(self):
        """health_check() runs @health decorated methods."""
        container = init(modules=[__name__])
        try:
            container.get(HealthyService)

            results = container.health_check()

            # Should have at least one health check result
            assert len(results) > 0
            # The HealthyService.is_healthy should return True
            healthy_checks = {k: v for k, v in results.items() if "HealthyService" in k and "Unhealthy" not in k}
            assert all(v is True for v in healthy_checks.values())
        finally:
            container.shutdown()

    def test_health_check_catches_exceptions(self):
        """health_check() returns False for failing checks."""
        container = init(modules=[__name__])
        try:
            container.get(UnhealthyService)

            results = container.health_check()

            # Should have a False entry for unhealthy service
            unhealthy_results = {k: v for k, v in results.items() if "Unhealthy" in k}
            assert False in unhealthy_results.values()
        finally:
            container.shutdown()


class TestContainerInfo:
    """Test container info logging."""

    def test_info_logs_with_container_id(self, caplog):
        """info() includes container ID prefix."""
        import logging

        container = init(modules=[__name__])
        try:
            with caplog.at_level(logging.INFO, logger="pico_ioc"):
                container.info("Test message")

            assert any("Test message" in r.message for r in caplog.records)
        finally:
            container.shutdown()



import importlib
import pkgutil
import logging
import inspect
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union
from .exceptions import ConfigurationError, InvalidBindingError
from .factory import ComponentFactory, ProviderMetadata
from .locator import ComponentLocator
from .scope import ScopeManager, ScopedCaches
from .container import PicoContainer
from .decorators import component, factory, provides, Qualifier, configure, cleanup, configured
from .constants import SCOPE_SINGLETON
from .config_builder import ContextConfig, configuration
from .registrar import Registrar
from .aop import ContainerObserver
from .component_scanner import CustomScanner

KeyT = Union[str, type]
Provider = Callable[[], Any]

def _scan_package(package) -> Iterable[Any]:
    for _, name, _ in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        yield importlib.import_module(name)

def _iter_input_modules(inputs: Union[Any, Iterable[Any]]) -> Iterable[Any]:
    seq = inputs if isinstance(inputs, Iterable) and not inspect.ismodule(inputs) and not isinstance(inputs, str) else [inputs]
    seen: Set[str] = set()
    for it in seq:
        if isinstance(it, str):
            mod = importlib.import_module(it)
        else:
            mod = it
        if hasattr(mod, "__path__"):
            for sub in _scan_package(mod):
                name = getattr(sub, "__name__", None)
                if name and name not in seen:
                    seen.add(name)
                    yield sub
        else:
            name = getattr(mod, "__name__", None)
            if name and name not in seen:
                seen.add(name)
                yield mod

def _normalize_override_provider(v: Any):
    if isinstance(v, tuple) and len(v) == 2:
        src, lz = v
        if callable(src):
            return (lambda s=src: s()), bool(lz)
        return (lambda s=src: s), bool(lz)
    if callable(v):
        return (lambda f=v: f()), False
    return (lambda inst=v: inst), False

def init(
    modules: Union[Any, Iterable[Any]],
    *,
    profiles: Tuple[str, ...] = (),
    allowed_profiles: Optional[Iterable[str]] = None,
    environ: Optional[Dict[str, str]] = None,
    overrides: Optional[Dict[KeyT, Any]] = None,
    logger: Optional[logging.Logger] = None,
    config: Optional[ContextConfig] = None,
    custom_scopes: Optional[Iterable[str]] = None,
    validate_only: bool = False,
    container_id: Optional[str] = None,
    observers: Optional[List[ContainerObserver]] = None,
    custom_scanners: Optional[List[CustomScanner]] = None,
) -> PicoContainer:
    active = tuple(p.strip() for p in profiles if p)
    
    allowed_set = set(a.strip() for a in allowed_profiles) if allowed_profiles is not None else None
    if allowed_set is not None:
        unknown = set(active) - allowed_set
        if unknown:
            raise ConfigurationError(f"Unknown profiles: {sorted(unknown)}; allowed: {sorted(allowed_set)}")
            
    factory = ComponentFactory()
    caches = ScopedCaches()
    scopes = ScopeManager()
    if custom_scopes:
        for name in custom_scopes:
            scopes.register_scope(name)
            
    pico = PicoContainer(factory, caches, scopes, container_id=container_id, profiles=active, observers=observers or [])
    registrar = Registrar(factory, profiles=active, environ=environ, logger=logger, config=config)
    
    if custom_scanners:
        for scanner in custom_scanners:
            registrar.register_custom_scanner(scanner)

    for m in _iter_input_modules(modules):
        registrar.register_module(m)
        
    if overrides:
        for k, v in overrides.items():
            prov, _ = _normalize_override_provider(v)
            factory.bind(k, prov)
            
    registrar.finalize(overrides, pico_instance=pico)
    if validate_only:
        locator = registrar.locator()
        pico.attach_locator(locator)
        _fail_fast_cycle_check(pico)
        return pico
        
    locator = registrar.locator()
    registrar.attach_runtime(pico, locator)
    pico.attach_locator(locator)
    _fail_fast_cycle_check(pico)
    
    if not validate_only:
        eager_singletons = []
        for key, md in locator._metadata.items():
            if md.scope == SCOPE_SINGLETON and not md.lazy:
                cache = pico._cache_for(key)
                instance = cache.get(key)
                if instance is None:
                    instance = pico.get(key)
                    eager_singletons.append(instance)
                else:
                    eager_singletons.append(instance)
                    
        configure_awaitables = []
        for instance in eager_singletons:
            res = pico._run_configure_methods(instance)
            if inspect.isawaitable(res):
                configure_awaitables.append(res)
                
        if configure_awaitables:
            raise ConfigurationError(
                "Sync init() found eagerly loaded singletons with async @configure methods. "
                "This can be caused by an async __ainit__ or async @configure. "
                "Use an async main function and await pico.aget() for those components, "
                "or mark them as lazy=True."
            )

    return pico

def _find_cycle(graph: Dict[KeyT, Tuple[KeyT, ...]]) -> Optional[Tuple[KeyT, ...]]:
    temp: Set[KeyT] = set()
    perm: Set[KeyT] = set()
    stack: List[KeyT] = []
    def visit(n: KeyT) -> Optional[Tuple[KeyT, ...]]:
        if n in perm:
            return None
        if n in temp:
            try:
                idx = stack.index(n)
                return tuple(stack[idx:] + [n])
            except ValueError:
                return tuple([n, n])
        temp.add(n)
        stack.append(n)
        for m in graph.get(n, ()):
            c = visit(m)
            if c:
                return c
        stack.pop()
        temp.remove(n)
        perm.add(n)
        return None
    for node in graph.keys():
        c = visit(node)
        if c:
            return c
    return None

def _format_key(k: KeyT) -> str:
    return getattr(k, "__name__", str(k))

def _fail_fast_cycle_check(pico: "PicoContainer") -> None:
    graph = pico.build_resolution_graph()
    cyc = _find_cycle(graph)
    if not cyc:
        return
    path = " -> ".join(_format_key(k) for k in cyc)
    raise InvalidBindingError([f"Circular dependency detected: {path}"])



import pytest
import contextvars
from pico_ioc import PicoContainer, ComponentFactory, ScopedCaches, ScopeManager, health
from pico_ioc.exceptions import AsyncResolutionError
from pico_ioc.locator import ComponentLocator

def test_container_ashutdown():
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    
    async def run():
        await c.ashutdown()
        assert c.container_id not in PicoContainer.all_containers()
        
    import asyncio
    asyncio.run(run())

def test_export_graph_no_locator():
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    with pytest.raises(RuntimeError, match="No locator attached"):
        c.export_graph("out.dot")

def test_resolve_args_edge_cases():
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    assert c._resolve_args(()) == {}
    
    locator = ComponentLocator({}, {})
    c.attach_locator(locator)
    assert c._resolve_args(()) == {}

def test_info_logging(caplog):
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    with caplog.at_level("INFO"):
        c.info("test message")
    assert "test message" in caplog.text

def test_scope_context_manager():
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    
    with c.scope("request", "req-1"):
        assert c.scopes.get_id("request") == "req-1"
    
    assert c.scopes.get_id("request") is None

def test_health_check_exception():
    c = PicoContainer(ComponentFactory(), ScopedCaches(), ScopeManager())
    
    class Unhealthy:
        # FIX: Use decorator
        @health
        def check(self):
            raise ValueError("fail")
            
    obj = Unhealthy()
    c._caches._singleton.put("key", obj)
    
    res = c.health_check()
    assert res["key.check"] is False



# tests/test_provides_module_functions.py
import types
import pytest
from pico_ioc.api import component, provides, init
from pico_ioc.exceptions import ProviderNotFoundError, InvalidBindingError

class Service:
    pass

class Dep:
    pass

class Impl(Service):
    def __init__(self, dep: Dep) -> None:
        self.dep = dep

def build_module_with_module_level_provides():
    m = types.ModuleType("module_level_provides")

    @component
    class MyDep(Dep):
        pass

    @provides(Service)
    def build_service(dep: Dep) -> Service:
        return Impl(dep)

    setattr(m, "MyDep", MyDep)
    setattr(m, "build_service", build_service)
    return m

def build_module_with_string_key():
    m = types.ModuleType("module_level_string_key")

    @provides("cache")
    def build_cache() -> dict:
        return {"ok": True}

    setattr(m, "build_cache", build_cache)
    return m

def build_module_with_profiles():
    m = types.ModuleType("module_level_profiles")

    @provides(Service, conditional_profiles=("prod",))
    def build_service_prod_only() -> Service:
        return Impl(Dep())

    setattr(m, "build_service_prod_only", build_service_prod_only)
    return m

def build_module_with_missing_dependency():
    m = types.ModuleType("module_level_missing_dep")

    class Missing:
        pass

    @provides(Service)
    def build_service(missing: "Missing") -> Service:
        return Impl(Dep())

    setattr(m, "Missing", Missing)
    setattr(m, "build_service", build_service)
    return m

def test_module_level_provides_binds_and_injects_dependency():
    mod = build_module_with_module_level_provides()
    pico = init(mod)
    s = pico.get(Service)
    assert isinstance(s, Impl)
    assert isinstance(s.dep, Dep)

def test_module_level_provides_with_string_key_and_retrieval_by_key():
    mod = build_module_with_string_key()
    pico = init(mod)
    c = pico.get("cache")
    assert isinstance(c, dict)
    assert c.get("ok") is True

def test_module_level_provides_respects_profiles_enabled():
    mod = build_module_with_profiles()
    pico = init(mod, profiles=("prod",))
    s = pico.get(Service)
    assert isinstance(s, Service)

def test_module_level_provides_respects_profiles_disabled():
    mod = build_module_with_profiles()
    pico = init(mod, profiles=("dev",))
    with pytest.raises(ProviderNotFoundError):
        pico.get(Service)

def test_module_level_provides_validation_reports_missing_dependencies():
    mod = build_module_with_missing_dependency()
    with pytest.raises(InvalidBindingError):
        init(mod, validate_only=True)




from .constants import LOGGER_NAME, LOGGER, PICO_INFRA, PICO_NAME, PICO_KEY, PICO_META
from .exceptions import (
    PicoError,
    ProviderNotFoundError,
    ComponentCreationError,
    ScopeError,
    ConfigurationError,
    SerializationError,
    ValidationError,
    InvalidBindingError,
    EventBusClosedError,
    AsyncResolutionError,
)
from .api import (
    component,
    factory,
    provides,
    Qualifier,
    configure,
    cleanup,
    init,
    configured,
)
from .config_builder import configuration, ContextConfig, EnvSource, FileSource, FlatDictSource, Value
from .scope import ScopeManager, ContextVarScope, ScopeProtocol, ScopedCaches
from .locator import ComponentLocator
from .factory import ComponentFactory, ProviderMetadata, DeferredProvider
from .aop import MethodCtx, MethodInterceptor, intercepted_by, UnifiedComponentProxy, health, ContainerObserver
from .container import PicoContainer
from .event_bus import EventBus, ExecPolicy, ErrorPolicy, Event, subscribe, AutoSubscriberMixin
from .config_runtime import JsonTreeSource, YamlTreeSource, DictSource, Discriminator, Value
from .analysis import DependencyRequest, analyze_callable_dependencies
from .component_scanner import CustomScanner

__all__ = [
    "LOGGER_NAME",
    "LOGGER",
    "PICO_INFRA",
    "PICO_NAME",
    "PICO_KEY",
    "PICO_META",
    "PicoError",
    "ProviderNotFoundError",
    "ComponentCreationError",
    "ScopeError",
    "ConfigurationError",
    "SerializationError",
    "ValidationError",
    "InvalidBindingError",
    "AsyncResolutionError",
    "EventBusClosedError",
    "component",
    "factory",
    "provides",
    "Qualifier",
    "configure",
    "cleanup",
    "ScopeProtocol",
    "ContextVarScope",
    "ScopeManager",
    "ComponentLocator",
    "ScopedCaches",
    "ProviderMetadata",
    "ComponentFactory",
    "DeferredProvider",
    "MethodCtx",
    "MethodInterceptor",
    "intercepted_by",
    "UnifiedComponentProxy",
    "health",
    "ContainerObserver",
    "PicoContainer",
    "EnvSource",
    "FileSource",
    "FlatDictSource",
    "init",
    "configured",
    "configuration",
    "ContextConfig",
    "Value",
    "EventBus",
    "ExecPolicy",
    "ErrorPolicy",
    "Event",
    "subscribe",
    "AutoSubscriberMixin",
    "JsonTreeSource",
    "YamlTreeSource",
    "DictSource",
    "Discriminator",
    "Value",
    "DependencyRequest",
    "analyze_callable_dependencies",
    "CustomScanner",
]



# tests/test_provides_static_methods.py
import types
from pico_ioc.api import component, factory, provides, init

class Service:
    pass

class Dep:
    pass

class Impl(Service):
    def __init__(self, dep: Dep) -> None:
        self.dep = dep

def build_module_with_staticmethod_provides():
    m = types.ModuleType("factory_staticmethod_provides")

    @component
    class MyDep(Dep):
        pass

    @factory
    class MyFactory:
        @staticmethod
        @provides(Service)
        def build(dep: Dep) -> Service:
            return Impl(dep)

    setattr(m, "MyDep", MyDep)
    setattr(m, "MyFactory", MyFactory)
    return m

def test_factory_staticmethod_provides_binds_and_injects():
    mod = build_module_with_staticmethod_provides()
    pico = init(mod)
    s = pico.get(Service)
    assert isinstance(s, Impl)
    assert isinstance(s.dep, Dep)




import pytest

from pico_ioc.aop import UnifiedComponentProxy


MockContainer = object()


class Tracker:
    created = 0

    def __init__(self):
        type(self).created += 1
        self.attr = "init"
        self.calls = 0
        self.entered = False
        self.exited = False
        self.store = {}

    def __call__(self, x):
        self.calls += 1
        return f"called:{x}"

    def __enter__(self):
        self.entered = True
        return self

    def __exit__(self, exc_type, exc, tb):
        self.exited = True
        return False


class Mat:
    def __init__(self, val):
        self.val = val

    def __matmul__(self, other):
        right = other.val if hasattr(other, "val") else other
        return ("matmul", self.val, right)

    def __rmatmul__(self, other):
        left = other.val if hasattr(other, "val") else other
        return ("rmatmul", left, self.val)


def test_lazy_instantiation_on_first_attribute_access():
    Tracker.created = 0

    def factory():
        return Tracker()

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)
    assert Tracker.created == 0

    assert p.attr == "init"
    assert Tracker.created == 1

    p.attr = "changed"
    assert p.attr == "changed"
    assert Tracker.created == 1


def test_dunder_len_iter_contains_with_list():
    created = {"n": 0}

    def factory():
        created["n"] += 1
        return [1, 2, 3]

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)
    assert created["n"] == 0

    assert len(p) == 3
    assert list(iter(p)) == [1, 2, 3]
    assert 2 in p
    assert created["n"] == 1

    assert list(reversed(p)) == [3, 2, 1]
    assert created["n"] == 1


def test_item_get_set_del_with_dict():
    created = {"n": 0}

    def factory():
        created["n"] += 1
        return {"a": 1}

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)
    assert created["n"] == 0

    assert p["a"] == 1
    p["b"] = 2
    assert p["b"] == 2
    del p["a"]
    assert "a" not in p
    assert created["n"] == 1


def test_math_ops_and_reflected_ops_with_int():
    made = {"n": 0}

    def factory():
        made["n"] += 1
        return 10

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert p + 5 == 15
    assert p - 3 == 7
    assert p * 2 == 20
    assert p // 3 == 3
    assert p % 4 == 2
    assert p ** 2 == 100

    assert 5 + p == 15
    assert 30 - p == 20
    assert 2 * p == 20

    assert hash(p) == hash(10)
    assert bool(p) is True
    assert p == 10
    assert made["n"] == 1


def test_call_delegation_and_repr_str():
    Tracker.created = 0

    def factory():
        return Tracker()

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert str(p) == str(p._get_real_object())
    assert repr(p) == repr(p._get_real_object())

    out = p("X")
    assert out == "called:X"
    assert p.calls == 1
    assert Tracker.created == 1


def test_context_manager_delegation():
    Tracker.created = 0

    def factory():
        return Tracker()

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    with p as obj:
        assert obj.entered is True
        obj.attr = "inside"

    assert p.exited is True
    assert p.attr == "inside"
    assert Tracker.created == 1


def test_dir_and_getattr_delattr_passthrough():
    def factory():
        return Tracker()

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    names = dir(p)
    assert "attr" in names and "__class__" in names

    assert getattr(p, "attr") == "init"
    setattr(p, "attr", "new")
    assert p.attr == "new"
    delattr(p, "attr")
    with pytest.raises(AttributeError):
        _ = p.attr

def test___class___property_mirrors_real_class_and_affects_isinstance():
    def factory():
        return Tracker()

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    from pico_ioc.aop import UnifiedComponentProxy as _ProxyType
    assert type(p) is _ProxyType

    assert p.__class__ is Tracker

    assert isinstance(p, Tracker) is True


def test_true_div_and_reflected_and_divmod():
    calls = {"n": 0}
    def factory():
        calls["n"] += 1
        return 10

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert p / 4 == pytest.approx(2.5)
    assert 20 / p == pytest.approx(2.0)

    assert divmod(p, 6) == (1, 4)
    assert divmod(23, p) == (2, 3)

    assert calls["n"] == 1


def test_bitwise_ops_and_reflected():
    calls = {"n": 0}
    def factory():
        calls["n"] += 1
        return 0b0110

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert (p & 0b0011) == 0b0010
    assert (p | 0b1000) == 0b1110
    assert (p ^ 0b0101) == 0b0011

    assert (0b0011 & p) == 0b0010
    assert (0b1000 | p) == 0b1110
    assert (0b0101 ^ p) == 0b0011

    assert calls["n"] == 1


def test_shift_ops_and_reflected():
    calls = {"n": 0}
    def factory():
        calls["n"] += 1
        return 3

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert (p << 2) == 12
    assert (p >> 1) == 1

    class Shiftable:
        def __init__(self, v): self.v = v
        def __lshift__(self, other): return ("L", self.v, other)
        def __rshift__(self, other): return ("R", self.v, other)

    s = Shiftable(5)
    assert (s << p) == ("L", 5, 3)
    assert (s >> p) == ("R", 5, 3)

    assert calls["n"] == 1


def test_unary_ops_and_ordered_comparisons():
    calls = {"n": 0}
    def factory():
        calls["n"] += 1
        return -3

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    assert -p == 3
    assert +p == -3
    assert abs(p) == 3
    assert ~p == 2

    assert p < -2
    assert p <= -3
    assert p > -4
    assert p >= -3

    assert calls["n"] == 1


def test_matmul_and_reflected_matmul():
    calls = {"n": 0}
    def factory():
        calls["n"] += 1
        return Mat(2)

    p = UnifiedComponentProxy(container=MockContainer, object_creator=factory)

    out1 = p @ 5
    assert out1 == ("matmul", 2, 5)

    out2 = 7 @ p
    assert out2 == ("rmatmul", 7, 2)

    other = Mat(9)
    assert (p @ other) == ("matmul", 2, 9)
    assert (other @ p) == ("matmul", 9, 2)

    assert calls["n"] == 1
